# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for annotation ingestion."""

import os
import shutil
import tempfile

from etils import epath
from ml_collections import config_dict
import numpy as np
from perch_hoplite import path_utils
from perch_hoplite.agile import embed
from perch_hoplite.agile import ingest_annotations
from perch_hoplite.agile.tests import test_utils
from perch_hoplite.db import datatypes
from perch_hoplite.db import interface
from perch_hoplite.db.tests import test_utils as db_test_utils
from perch_hoplite.taxonomy import annotations
from perch_hoplite.taxonomy import annotations_fns

from absl.testing import absltest


class IngestAnnotationsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    # `self.create_tempdir()` raises an UnparsedFlagAccessError, which is why
    # we use `tempdir` directly.
    self.tempdir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  def make_annotated_data(self):
    sites = ['site_1', 'site_2', 'site_3']
    filenames = ['foo', 'bar', 'baz']
    test_utils.make_wav_files(self.tempdir, sites, filenames, file_len_s=60.0)
    classes = ['x', 'y', 'z']

    # Make a collection of random annotations.
    annos = []
    for i, s in enumerate(sites):
      for j, f in enumerate(filenames):
        for k, c in enumerate(classes):
          annos.append(
              annotations.TimeWindowAnnotation(
                  filename=f'{s}/{f}_{s}.wav',
                  start_time_s=9 * i + 3 * j + k,
                  end_time_s=9 * i + 3 * j + k + 1,
                  namespace='test',
                  label=[c],
              )
          )

    annos_path = os.path.join(self.tempdir, 'annos.csv')
    annotations.write_annotations_csv(annos_path, annos)
    return annos_path, annos

  def test_embed_and_ingest_annotations(self):
    rng = np.random.default_rng(42)
    db = db_test_utils.make_db(
        self.tempdir,
        'in_mem',
        num_embeddings=1000,
        rng=rng,
        fill_random=False,
    )
    deployment_id = db.insert_deployment(name='hawaii', project='hawaii')
    recording_id = db.insert_recording(
        filename='UHH_001_S01_20161121_150000.flac',
        deployment_id=deployment_id,
    )
    window_size_s = 5.0
    emb_offsets = [175, 185, 275, 230, 235]
    windows = []
    for offset in emb_offsets:
      emb_idx = db.insert_window(
          recording_id=recording_id,
          offsets=[offset, offset + window_size_s],
          embedding=rng.normal([db.get_embedding_dim()]),
      )
      windows.append(db.get_window(emb_idx))

    hawaii_annos_path = path_utils.get_absolute_path(
        'agile/tests/testdata/hawaii.csv'
    )
    ingestor = ingest_annotations.AnnotatedDatasetIngestor(
        base_path=hawaii_annos_path.parent,
        audio_glob='*/*.flac',
        dataset_name='hawaii',
        annotation_filename='hawaii.csv',
        annotation_load_fn=annotations_fns.load_cornell_annotations,
    )
    inserted_labels = ingestor.ingest_dataset(db, provenance='test_dataset')
    self.assertSetEqual(
        set(inserted_labels.keys()), {'jabwar', 'hawama', 'ercfra'}
    )
    # Check that individual labels are correctly applied.
    # The Hawai'i test data CSV contains a total of five annotations.
    # The window at offset 175 should have no labels.
    self.assertEmpty(
        db.get_all_annotations(
            config_dict.create(
                eq=dict(recording_id=windows[0].recording_id),
                approx=dict(offsets=windows[0].offsets),
            )
        )
    )  # offset 175

    def _check_label(want_label_str, got_label):
      self.assertEqual(got_label.label, want_label_str)
      self.assertEqual(got_label.label_type, datatypes.LabelType.POSITIVE)
      self.assertEqual(got_label.provenance, 'test_dataset')

    # There are two jabwar annotations for the window at offset 185.
    offset_185_labels = db.get_all_annotations(
        config_dict.create(
            eq=dict(recording_id=windows[1].recording_id),
            approx=dict(offsets=windows[1].offsets),
        )
    )
    self.assertLen(offset_185_labels, 2)
    _check_label('jabwar', offset_185_labels[0])
    _check_label('jabwar', offset_185_labels[1])

    offset_275_labels = db.get_all_annotations(
        config_dict.create(
            eq=dict(recording_id=windows[2].recording_id),
            approx=dict(offsets=windows[2].offsets),
        )
    )
    self.assertLen(offset_275_labels, 1)
    _check_label('hawama', offset_275_labels[0])

    self.assertEmpty(
        db.get_all_annotations(
            config_dict.create(
                eq=dict(recording_id=windows[3].recording_id),
                approx=dict(offsets=windows[3].offsets),
            )
        )
    )  # offset 230

    offset_235_labels = db.get_all_annotations(
        config_dict.create(
            eq=dict(recording_id=windows[4].recording_id),
            approx=dict(offsets=windows[4].offsets),
        )
    )
    self.assertLen(offset_235_labels, 1)
    _check_label('ercfra', offset_235_labels[0])

  def test_ingest_annotations(self):
    annos_path, annos = self.make_annotated_data()
    self.assertLen(annos, 27)

    def _loader_fn(x):
      annos = annotations.read_annotations_csv(x, namespace='somedata')
      return annotations.annotations_to_dataframe(annos)

    ingestor = ingest_annotations.AnnotatedDatasetIngestor(
        base_path=epath.Path(self.tempdir),
        audio_glob='*/*.wav',
        dataset_name='test',
        annotation_filename=annos_path,
        annotation_load_fn=_loader_fn,
    )
    placeholder_model_config = config_dict.ConfigDict()
    placeholder_model_config.embedding_size = 32
    placeholder_model_config.sample_rate = 16000
    model_config = embed.ModelConfig(
        model_key='placeholder_model',
        embedding_dim=32,
        model_config=placeholder_model_config,
    )
    db, ingestor_class_counts = ingest_annotations.embed_annotated_dataset(
        ds_choice=ingestor,
        db_path=os.path.join(self.tempdir),
        db_model_config=model_config,
    )
    self.assertEqual(db.count_embeddings(), 60 * 9)

    ingestor.ingest_dataset(db)
    class_counts = db.count_each_label()
    for lbl in ('x', 'y', 'z'):
      self.assertEqual(class_counts[lbl], 9)
      self.assertEqual(ingestor_class_counts[lbl], 9)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for manipulating annotations."""

import csv
import dataclasses
from typing import Callable, Sequence

from etils import epath
import pandas as pd


@dataclasses.dataclass
class TimeWindowAnnotation:
  """An annotation for a particular time window.

  Attributes:
    filename: Filename for the source audio.
    start_time_s: Float representing the start of this annotation window.
    end_time_s: Float representing the end of this annotation window.
    namespace: The namespace of the classes in this annotation.
    label: List of classes present in the audio segment.
  """

  filename: str
  start_time_s: float
  end_time_s: float
  namespace: str
  label: Sequence[str]


def annotations_to_dataframe(
    annotations: Sequence[TimeWindowAnnotation],
) -> pd.DataFrame:
  return pd.DataFrame.from_records(
      [dataclasses.asdict(anno) for anno in annotations]
  )


def write_annotations_csv(
    filepath: str | epath.Path,
    annotations: Sequence[TimeWindowAnnotation],
    label_separator: str = ' ',
) -> None:
  """Write annotations to a CSV file."""
  fieldnames = [f.name for f in dataclasses.fields(TimeWindowAnnotation)]
  fieldnames.remove('namespace')
  with epath.Path(filepath).open('w') as f:
    dr = csv.DictWriter(f, fieldnames)
    dr.writeheader()
    for anno in annotations:
      anno_dict = {f: getattr(anno, f) for f in fieldnames}
      anno_dict['label'] = label_separator.join(anno_dict['label'])
      dr.writerow(anno_dict)


def read_annotations_csv(
    annotations_filepath: epath.Path, namespace: str, label_separator: str = ' '
) -> Sequence[TimeWindowAnnotation]:
  """Read annotations as written by write_annotations_csv."""
  got_annotations = []
  with epath.Path(annotations_filepath).open('r') as f:
    dr = csv.DictReader(f)
    for row in dr:
      got_annotations.append(
          TimeWindowAnnotation(
              filename=row['filename'],
              namespace=namespace,
              start_time_s=float(row['start_time_s']),
              end_time_s=float(row['end_time_s']),
              label=row['label'].split(label_separator),
          )
      )
  return got_annotations


def read_dataset_annotations_csvs(
    filepaths: Sequence[epath.Path],
    filename_fn: Callable[[epath.Path, dict[str, str]], str],
    namespace: str,
    class_fn: Callable[[dict[str, str]], Sequence[str]],
    start_time_fn: Callable[[dict[str, str]], float],
    end_time_fn: Callable[[dict[str, str]], float],
    filter_fn: Callable[[dict[str, str]], bool] | None = None,
    delimiter: str = ',',
) -> Sequence[TimeWindowAnnotation]:
  """Create annotations from a random CSV.

  Args:
    filepaths: Path to the CSV files.
    filename_fn: Function for extracting the audio filename. Maps
      (annotations_filename, row) to the filename of the audio.
    namespace: Namespace for the annotated classes.
    class_fn: Function for extracting classname.
    start_time_fn: Field for starting timestamps. Currently assumes values are
      floats measured in seconds.
    end_time_fn: Field for ending timestamps.
    filter_fn: A function for selecting rows of the annotation file to ignore.
      Will keep rows where filter_fn is False, and ignore rows where True.
    delimiter: Field separating character in the target file.

  Returns:
    List of TimeWindowAnnotations.
  """
  annotations = []
  for filepath in filepaths:
    with filepath.open('r') as f:
      reader = csv.DictReader(f, delimiter=delimiter)
      for row in reader:
        if filter_fn and filter_fn(row):
          continue
        filename = filename_fn(filepath, row)
        start = start_time_fn(row)
        end = end_time_fn(row)
        classes = class_fn(row)
        annotations.append(
            TimeWindowAnnotation(filename, start, end, namespace, classes)
        )
  return annotations



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of a MultiDB wrapper for HopliteDBInterface."""

import collections
from collections.abc import Sequence
import datetime as dt
from typing import Any, Literal

from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import datatypes
from perch_hoplite.db import interface


class MultiDBWrapper(interface.HopliteDBInterface):
  """A wrapper around multiple HopliteDBInterface instances.

  ID numbers are translated internally to avoid collisions.
  """

  def __init__(self, dbs: Sequence[interface.HopliteDBInterface]):
    """Initialize with multiple databases.

    Args:
      dbs: A sequence of databases to wrap.

    Raises:
      ValueError: If `MultiDBWrapper` contains no databases.
    """
    self._dbs = dbs
    if not self._dbs:
      raise ValueError("MultiDBWrapper contains no databases.")

  def _split_id(self, external_id: int) -> tuple[int, int]:
    """Splits an external ID into a database index and internal ID.

    Calculate the internal database index and internal ID from an external ID
    number.

    Args:
      external_id: The external ID number.

    Returns:
      db_index: The database index.
      internal_id: The internal ID number within the database that has index
        db_index.
    """

    num_dbs = len(self._dbs)
    db_index = external_id % num_dbs
    internal_id = external_id // num_dbs
    self._dbs[db_index].get_window(internal_id)
    return db_index, internal_id

  def _join_id(self, db_index: int, internal_id: int) -> int:
    """Joins a database index and internal ID into an external ID.

    Args:
      db_index: The database index.
      internal_id: The internal ID number.

    Returns:
      The external ID number.
    """

    self._dbs[db_index].get_window(internal_id)
    return internal_id * len(self._dbs) + db_index

  @classmethod
  def create(cls, **kwargs) -> "MultiDBWrapper":
    raise NotImplementedError()

  def add_extra_table_column(
      self,
      table_name: str,
      column_name: str,
      column_type: type[Any],
  ) -> None:
    raise NotImplementedError()

  def get_extra_table_columns(self) -> dict[str, dict[str, type[Any]]]:
    raise NotImplementedError()

  def commit(self) -> None:
    raise NotImplementedError()

  def rollback(self) -> None:
    raise NotImplementedError()

  def thread_split(self) -> "MultiDBWrapper":
    raise NotImplementedError()

  def insert_metadata(self, key: str, value: config_dict.ConfigDict) -> None:
    raise NotImplementedError()

  def get_metadata(self, key: str | None) -> config_dict.ConfigDict:
    raise NotImplementedError()

  def remove_metadata(self, key: str | None) -> None:
    raise NotImplementedError()

  def insert_deployment(
      self,
      name: str,
      project: str,
      latitude: float | None = None,
      longitude: float | None = None,
      **kwargs: Any,
  ) -> int:
    raise NotImplementedError()

  def get_deployment(self, deployment_id: int) -> datatypes.Deployment:
    raise NotImplementedError()

  def remove_deployment(self, deployment_id: int) -> None:
    raise NotImplementedError()

  def insert_recording(
      self,
      filename: str,
      datetime: dt.datetime | None = None,
      deployment_id: int | None = None,
      **kwargs: Any,
  ) -> int:
    raise NotImplementedError()

  def get_recording(self, recording_id: int) -> datatypes.Recording:
    raise NotImplementedError()

  def remove_recording(self, recording_id: int) -> None:
    raise NotImplementedError()

  def insert_window(
      self,
      recording_id: int,
      offsets: list[float],
      embedding: np.ndarray | None = None,
      handle_duplicates: Literal[
          "allow", "overwrite", "skip", "error"
      ] = "error",
      **kwargs: Any,
  ) -> int:
    raise NotImplementedError()

  def get_window(
      self,
      window_id: int,
      include_embedding: bool = False,
  ) -> datatypes.Window:
    raise NotImplementedError()

  def get_embedding(self, window_id: int) -> np.ndarray:
    raise NotImplementedError()

  def remove_window(self, window_id: int) -> None:
    raise NotImplementedError()

  def insert_annotation(
      self,
      recording_id: int,
      offsets: list[float],
      label: str,
      label_type: datatypes.LabelType,
      provenance: str,
      handle_duplicates: Literal[
          "allow", "overwrite", "skip", "error"
      ] = "error",
      **kwargs: Any,
  ) -> int:
    raise NotImplementedError()

  def get_annotation(self, annotation_id: int) -> datatypes.Annotation:
    raise NotImplementedError()

  def remove_annotation(self, annotation_id: int) -> None:
    raise NotImplementedError()

  def count_embeddings(self) -> int:
    raise NotImplementedError()

  def match_window_ids(
      self,
      deployments_filter: config_dict.ConfigDict | None = None,
      recordings_filter: config_dict.ConfigDict | None = None,
      windows_filter: config_dict.ConfigDict | None = None,
      annotations_filter: config_dict.ConfigDict | None = None,
      limit: int | None = None,
  ) -> Sequence[int]:
    raise NotImplementedError()

  def get_all_projects(self) -> Sequence[str]:
    raise NotImplementedError()

  def get_all_deployments(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Deployment]:
    raise NotImplementedError()

  def get_all_recordings(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Recording]:
    raise NotImplementedError()

  def get_all_windows(
      self,
      include_embedding: bool = False,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Window]:
    raise NotImplementedError()

  def get_all_annotations(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Annotation]:
    raise NotImplementedError()

  def get_all_labels(
      self,
      label_type: datatypes.LabelType | None = None,
  ) -> Sequence[str]:
    raise NotImplementedError()

  def count_each_label(
      self,
      label_type: datatypes.LabelType | None = None,
  ) -> collections.Counter[str]:
    raise NotImplementedError()

  def get_embedding_dim(self) -> int:
    raise NotImplementedError()

  def get_embedding_dtype(self) -> type[Any]:
    raise NotImplementedError()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""In-memory database implementation."""

import collections
from collections.abc import Sequence
import copy
import dataclasses
import datetime as dt
import itertools
from typing import Any, Literal

from absl import logging
from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import datatypes
from perch_hoplite.db import interface


def select_matching_keys(
    kv: dict[int, Any],
    filter_dict: config_dict.ConfigDict | None = None,
) -> set[int]:
  """Select the keys from a dictionary whose values match the given constraints.

  Args:
    kv: The dictionary to verify against the constraints.
    filter_dict: An optional ConfigDict of constraints to verify.

  Returns:
    The keys of the matching items.
  """

  if not filter_dict:
    return set(kv.keys())

  supported_ops = {
      'eq',
      'neq',
      'lt',
      'lte',
      'gt',
      'gte',
      'isin',
      'notin',
      'range',
      'approx',
  }
  for op_name, op_filters in filter_dict.items():
    if op_name not in supported_ops:
      raise ValueError(
          f'Unsupported operation: `{op_name}`. Supported filtering operations'
          f' are: {supported_ops}.'
      )
    if not isinstance(op_filters, config_dict.ConfigDict):
      raise ValueError(f'`{op_name}` value must be a ConfigDict.')

  def _is_match(obj: Any) -> bool:
    for op_name, op_filters in filter_dict.items():
      for key, value in op_filters.items():
        attr = getattr(obj, key, None)

        if op_name == 'eq':
          if key == 'offsets':
            logging.warning(
                "Do not apply `eq` to the `offsets` unless you know what you're"
                ' doing. Apply `approx` instead to avoid floating point errors.'
            )
          if attr is None:
            if value is not None:
              return False
          else:
            if isinstance(attr, np.ndarray):
              if (attr != value).any():
                return False
            else:
              if attr != value:
                return False
        elif op_name == 'neq':
          if attr is None:
            if value is None:
              return False
          else:
            if isinstance(attr, np.ndarray):
              if (attr != value).all():
                return False
            else:
              if attr == value:
                return False
        elif op_name == 'lt':
          if attr is None or value is None:
            return False
          else:
            if attr >= value:
              return False
        elif op_name == 'lte':
          if attr is None or value is None:
            return False
          else:
            if attr > value:
              return False
        elif op_name == 'gt':
          if attr is None or value is None:
            return False
          else:
            if attr <= value:
              return False
        elif op_name == 'gte':
          if attr is None or value is None:
            return False
          else:
            if attr < value:
              return False
        elif op_name == 'isin':
          if not isinstance(value, list):
            raise ValueError(f'`{op_name}` value must be a list.')
          if attr not in value:
            return False
        elif op_name == 'notin':
          if not isinstance(value, list):
            raise ValueError(f'`{op_name}` value must be a list.')
          if attr in value:
            return False
        elif op_name == 'range':
          if not isinstance(value, list) or len(value) != 2:
            raise ValueError(f'`{op_name}` value must be a list of 2 elements.')
          if attr is None:
            return False
          if attr < value[0] or attr > value[1]:
            return False
        elif op_name == 'approx':
          if attr is None or value is None:
            return False
          if key == 'offsets':
            if not np.allclose(attr, value, rtol=0.0, atol=1e-6):
              return False
          else:
            if abs(attr - value) >= 1e-6:
              return False

    return True

  return {key for key, value in kv.items() if _is_match(value)}


@dataclasses.dataclass
class InMemoryGraphSearchDB(interface.HopliteDBInterface):
  """In-memory hoplite database."""

  # User-provided.
  _embedding_dim: int
  _embedding_dtype: type[Any]

  # Dynamic state.
  _extra_table_columns: dict[str, dict[str, type[Any]]] = dataclasses.field(
      default_factory=lambda: {
          'deployments': {},
          'recordings': {},
          'windows': {},
          'annotations': {},
      }
  )

  # Storage for metadata & embeddings.
  _hoplite_metadata: dict[str, config_dict.ConfigDict] = dataclasses.field(
      default_factory=dict
  )
  _deployments: dict[int, datatypes.Deployment] = dataclasses.field(
      default_factory=dict
  )
  _recordings: dict[int, datatypes.Recording] = dataclasses.field(
      default_factory=dict
  )
  _windows: dict[int, datatypes.Window] = dataclasses.field(
      default_factory=dict
  )
  _annotations: dict[int, datatypes.Annotation] = dataclasses.field(
      default_factory=dict
  )
  _next_deployment_id: int = 1
  _next_recording_id: int = 1
  _next_window_id: int = 1
  _next_annotation_id: int = 1

  @classmethod
  def create(
      cls,
      embedding_dim: int,
      embedding_dtype: type[Any] = np.float16,
  ) -> 'InMemoryGraphSearchDB':
    """Connect to and, if needed, initialize the database."""
    return cls(
        _embedding_dim=embedding_dim,
        _embedding_dtype=embedding_dtype,
    )

  def add_extra_table_column(
      self,
      table_name: str,
      column_name: str,
      column_type: type[Any],
  ) -> None:
    """Add an extra column to a table in the database.."""
    self._extra_table_columns[table_name][column_name] = column_type

  def get_extra_table_columns(self) -> dict[str, dict[str, type[Any]]]:
    """Get all extra columns in the database."""
    return self._extra_table_columns

  def commit(self) -> None:
    """No-op to commit any pending transactions to the database.."""
    pass

  def rollback(self) -> None:
    """No-op to rollback any pending transactions to the database."""
    pass

  def thread_split(self) -> 'InMemoryGraphSearchDB':
    """Return the same database object since all data is in shared memory."""
    return self

  def insert_metadata(self, key: str, value: config_dict.ConfigDict) -> None:
    """Insert a key-value pair into the metadata table."""
    self._hoplite_metadata[key] = value

  def get_metadata(self, key: str | None) -> config_dict.ConfigDict:
    """Get a key-value pair from the metadata table."""

    if key is None:
      return config_dict.ConfigDict(self._hoplite_metadata)
    else:
      return self._hoplite_metadata[key]

  def remove_metadata(self, key: str | None) -> None:
    """Remove a key-value pair from the metadata table."""

    if key is None:
      self._hoplite_metadata.clear()
    else:
      del self._hoplite_metadata[key]

  def insert_deployment(
      self,
      name: str,
      project: str,
      latitude: float | None = None,
      longitude: float | None = None,
      **kwargs: Any,
  ) -> int:
    """Insert a deployment into the database."""

    deployment_id = self._next_deployment_id
    self._deployments[deployment_id] = datatypes.Deployment(
        id=deployment_id,
        name=name,
        project=project,
        latitude=latitude,
        longitude=longitude,
        **kwargs,
    )
    self._next_deployment_id += 1
    return deployment_id

  def get_deployment(self, deployment_id: int) -> datatypes.Deployment:
    """Get a deployment from the database."""
    deployment_id = int(deployment_id)
    return self._deployments[deployment_id]

  def remove_deployment(self, deployment_id: int) -> None:
    """Remove a deployment from the database."""

    deployment_id = int(deployment_id)

    remove_recording_ids = [
        recording.id
        for recording in self._recordings.values()
        if recording.deployment_id == deployment_id
    ]
    for recording_id in remove_recording_ids:
      self.remove_recording(recording_id)
    del self._deployments[deployment_id]

  def insert_recording(
      self,
      filename: str,
      datetime: dt.datetime | None = None,
      deployment_id: int | None = None,
      **kwargs: Any,
  ) -> int:
    """Insert a recording into the database."""

    if deployment_id is not None and deployment_id not in self._deployments:
      raise ValueError(f'Deployment id not found: {deployment_id}')

    recording_id = self._next_recording_id
    self._recordings[recording_id] = datatypes.Recording(
        id=recording_id,
        filename=filename,
        datetime=datetime,
        deployment_id=deployment_id,
        **kwargs,
    )
    self._next_recording_id += 1
    return recording_id

  def get_recording(
      self,
      recording_id: int,
  ) -> datatypes.Recording:
    """Get a recording from the database."""
    recording_id = int(recording_id)
    return self._recordings[recording_id]

  def remove_recording(self, recording_id: int) -> None:
    """Remove a recording from the database."""

    recording_id = int(recording_id)

    remove_window_ids = [
        window.id
        for window in self._windows.values()
        if window.recording_id == recording_id
    ]
    for window_id in remove_window_ids:
      self.remove_window(window_id)

    remove_annotation_ids = [
        annotation.id
        for annotation in self._annotations.values()
        if annotation.recording_id == recording_id
    ]
    for annotation_id in remove_annotation_ids:
      self.remove_annotation(annotation_id)

    del self._recordings[recording_id]

  def insert_window(
      self,
      recording_id: int,
      offsets: list[float],
      embedding: np.ndarray | None = None,
      handle_duplicates: Literal[
          'allow', 'overwrite', 'skip', 'error'
      ] = 'error',
      **kwargs: Any,
  ) -> int:
    """Insert a window into the database."""

    if recording_id not in self._recordings:
      raise ValueError(f'Recording id not found: {recording_id}')

    duplicate_id = self._handle_window_duplicates(
        recording_id, offsets, handle_duplicates
    )
    if duplicate_id is not None:
      return duplicate_id

    window_id = self._next_window_id
    self._windows[window_id] = datatypes.Window(
        id=window_id,
        recording_id=recording_id,
        offsets=offsets,
        embedding=embedding,
        **kwargs,
    )
    self._next_window_id += 1
    return window_id

  def get_window(
      self,
      window_id: int,
      include_embedding: bool = False,
  ) -> datatypes.Window:
    """Get a window from the database."""

    window_id = int(window_id)
    window = self._windows[window_id]

    if include_embedding:
      return window

    window = copy.copy(window)
    window.embedding = None
    return window

  def get_embedding(self, window_id: int) -> np.ndarray:
    """Get an embedding vector from the database."""
    window_id = int(window_id)
    return self._windows[window_id].embedding

  def remove_window(self, window_id: int) -> None:
    """Remove a window from the database."""
    window_id = int(window_id)
    del self._windows[window_id]

  def insert_annotation(
      self,
      recording_id: int,
      offsets: list[float],
      label: str,
      label_type: datatypes.LabelType,
      provenance: str,
      handle_duplicates: Literal[
          'allow', 'overwrite', 'skip', 'error'
      ] = 'error',
      **kwargs: Any,
  ) -> int:
    """Insert an annotation into the database."""

    if recording_id not in self._recordings:
      raise ValueError(f'Recording id not found: {recording_id}')

    duplicate_id = self._handle_annotation_duplicates(
        recording_id, offsets, label, label_type, provenance, handle_duplicates
    )
    if duplicate_id is not None:
      return duplicate_id

    annotation_id = self._next_annotation_id
    self._annotations[annotation_id] = datatypes.Annotation(
        id=annotation_id,
        recording_id=recording_id,
        offsets=offsets,
        label=label,
        label_type=label_type,
        provenance=provenance,
        **kwargs,
    )
    self._next_annotation_id += 1
    return annotation_id

  def get_annotation(self, annotation_id: int) -> datatypes.Annotation:
    """Get an annotation from the database."""
    annotation_id = int(annotation_id)
    return self._annotations[annotation_id]

  def remove_annotation(self, annotation_id: int) -> None:
    """Remove an annotation from the database."""
    annotation_id = int(annotation_id)
    del self._annotations[annotation_id]

  def count_embeddings(self) -> int:
    """Get the number of embeddings in the database."""
    return len(self._windows)

  def match_window_ids(
      self,
      deployments_filter: config_dict.ConfigDict | None = None,
      recordings_filter: config_dict.ConfigDict | None = None,
      windows_filter: config_dict.ConfigDict | None = None,
      annotations_filter: config_dict.ConfigDict | None = None,
      limit: int | None = None,
  ) -> Sequence[int]:
    """Get matching window IDs from the database based on given filters."""

    if limit is not None and limit <= 0:
      raise ValueError('Limit must be None or positive.')

    # Filter by deployment constraints.
    if deployments_filter:
      restrict_deployments = select_matching_keys(
          self._deployments,
          deployments_filter,
      )
    else:
      restrict_deployments = None

    # Filter by recording constraints.
    if recordings_filter or restrict_deployments is not None:
      if recordings_filter:
        restrict_recordings = select_matching_keys(
            self._recordings,
            recordings_filter,
        )
      else:
        restrict_recordings = set(self._recordings.keys())
      if restrict_deployments is not None:
        restrict_recordings &= {
            key
            for key, value in self._recordings.items()
            if value.deployment_id in restrict_deployments
        }
    else:
      restrict_recordings = None

    # Filter by window constraints.
    if windows_filter or restrict_recordings is not None:
      if windows_filter:
        restrict_windows = select_matching_keys(
            self._windows,
            windows_filter,
        )
      else:
        restrict_windows = set(self._windows.keys())
      if restrict_recordings is not None:
        restrict_windows &= {
            key
            for key, value in self._windows.items()
            if value.recording_id in restrict_recordings
        }
    else:
      restrict_windows = set(self._windows.keys())

    # Filter by annotation constraints.
    if annotations_filter:
      restrict_annotations = select_matching_keys(
          self._annotations,
          annotations_filter,
      )
      restrict_recording_offsets = {
          (
              self._annotations[annotation_id].recording_id,
              tuple(self._annotations[annotation_id].offsets),
          )
          for annotation_id in restrict_annotations
      }
      restrict_windows = {
          window_id
          for window_id in restrict_windows
          if (
              self._windows[window_id].recording_id,
              tuple(self._windows[window_id].offsets),
          )
          in restrict_recording_offsets
      }

    # Return the window IDs that match the constraints.
    if limit is None:
      return list(restrict_windows)
    return list(itertools.islice(restrict_windows, limit))

  def get_all_projects(self) -> Sequence[str]:
    """Get all distinct projects from the database."""
    return sorted({d.project for d in self._deployments.values()})

  def get_all_deployments(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Deployment]:
    """Get all deployments from the database."""
    restrict_deployments = select_matching_keys(self._deployments, filter)
    return [self._deployments[key] for key in restrict_deployments]

  def get_all_recordings(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Recording]:
    """Get all recordings from the database."""
    restrict_recordings = select_matching_keys(self._recordings, filter)
    return [self._recordings[key] for key in restrict_recordings]

  def get_all_windows(
      self,
      include_embedding: bool = False,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Window]:
    """Get all windows from the database."""

    restrict_windows = select_matching_keys(self._windows, filter)
    windows = [self._windows[key] for key in restrict_windows]

    if include_embedding:
      return windows

    windows = [copy.copy(window) for window in windows]
    for window in windows:
      window.embedding = None
    return windows

  def get_all_annotations(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Annotation]:
    """Get all annotations from the database."""
    restrict_annotations = select_matching_keys(self._annotations, filter)
    return [self._annotations[key] for key in restrict_annotations]

  def get_all_labels(
      self,
      label_type: datatypes.LabelType | None = None,
  ) -> Sequence[str]:
    """Get all distinct labels from the database."""
    return sorted({
        a.label
        for a in self._annotations.values()
        if label_type is None or a.label_type == label_type
    })

  def count_each_label(
      self,
      label_type: datatypes.LabelType | None = None,
  ) -> collections.Counter[str]:
    """Count each label in the database, ignoring provenance."""

    # Avoid double-counting the same label on the same recording offsets because
    # of different provenances.
    unique_annotations = {
        (a.recording_id, tuple(a.offsets), a.label, a.label_type)
        for a in self._annotations.values()
        if label_type is None or a.label_type == label_type
    }
    return collections.Counter([a[2] for a in unique_annotations])

  def get_embedding_dim(self) -> int:
    """Get the embedding dimension."""
    return self._embedding_dim

  def get_embedding_dtype(self) -> type[Any]:
    """Get the embedding data type."""
    return self._embedding_dtype



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Brute force search and reranking utilities."""

import concurrent
import threading
from typing import Any, cast

import numpy as np
from perch_hoplite.db import interface
from perch_hoplite.db import score_functions
from perch_hoplite.db import search_results


def worker_initializer(state):
  name = threading.current_thread().name
  state[name + 'db'] = state['db'].thread_split()


def brute_search_worker_fn(emb_ids: np.ndarray, state: dict[str, Any]):
  name = threading.current_thread().name
  embeddings = state[name + 'db'].get_embeddings_batch(emb_ids)
  scores = state['score_fn'](embeddings, state['query_embedding'])
  top_locs = np.argpartition(scores, state['search_list_size'], axis=-1)
  return emb_ids[top_locs], scores[top_locs]


def threaded_brute_search(
    db: interface.HopliteDBInterface,
    query_embedding: np.ndarray,
    search_list_size: int,
    score_fn: score_functions.ScoreFn,
    batch_size: int = 1024,
    max_workers: int = 8,
    sample_size: int | float | None = None,
    rng_seed: int | None = None,
) -> search_results.TopKSearchResults:
  """Performs a brute-force search for neighbors of the query embedding.

  Args:
    db: Graph DB instance.
    query_embedding: Query embedding vector.
    search_list_size: Number of results to return.
    score_fn: Scoring function to use for ranking results.
    batch_size: Number of embeddings to score in each thread.
    max_workers: Maximum number of threads to use for the search.
    sample_size: Number of embeddings to compare with. If None, all embeddings
      are used. For floats between 0 and 1, sample a proportion of the database.
      For ints, sample the specified number of embeddings.
    rng_seed: Random number generator seed to use for sampled search.

  Returns:
    A TopKSearchResults object containing the search results.
  """
  state = {}
  state['search_list_size'] = search_list_size
  state['db'] = db
  state['query_embedding'] = query_embedding
  state['score_fn'] = score_fn

  results = search_results.TopKSearchResults(search_list_size)
  # Commit the DB, since we are about to create views in multiple threads.
  db.commit()
  with concurrent.futures.ThreadPoolExecutor(
      max_workers=max_workers,
      initializer=worker_initializer,
      initargs=(state,),
  ) as executor:
    ids = get_brute_search_ids(db, sample_size, rng_seed)
    futures = []
    for q in range(0, ids.shape[0], batch_size):
      futures.append(
          executor.submit(
              brute_search_worker_fn, ids[q : q + batch_size], state
          )
      )
    for f in futures:
      idxes, scores = f.result()
      for idx, score in zip(idxes, scores):
        if not results.will_filter(idx, score):
          results.update(
              search_results.SearchResult(idx, score), force_insert=True
          )
  return results


def brute_search(
    db: interface.HopliteDBInterface,
    query_embedding: np.ndarray,
    search_list_size: int,
    score_fn: score_functions.ScoreFn,
    sample_size: int | float | None = None,
    rng_seed: int | None = None,
) -> search_results.TopKSearchResults:
  """Performs a brute-force search for neighbors of the query embedding.

  Args:
    db: Graph DB instance.
    query_embedding: Query embedding vector.
    search_list_size: Number of results to return.
    score_fn: Scoring function to use for ranking results.
    sample_size: Number of embeddings to compare with. If None, all embeddings
      are used. For floats between 0 and 1, sample a proportion of the database.
      For ints, sample the specified number of embeddings.
    rng_seed: Random number generator seed to use for sampled search.

  Returns:
    A TopKSearchResults object containing the search results.
  """
  results = search_results.TopKSearchResults(search_list_size)
  ids = get_brute_search_ids(db, sample_size, rng_seed)
  for idx in ids:
    target_embedding = db.get_embedding(idx)
    score = score_fn(query_embedding, target_embedding)
    # Check filtering and then force insert to avoid creating a SearchResult
    # object for discarded objects. This saves a small amount of time in the
    # inner loop.
    if not results.will_filter(idx, score):
      results.update(search_results.SearchResult(idx, score), force_insert=True)
  return results


def get_brute_search_ids(
    db: interface.HopliteDBInterface,
    sample_size: int | float | None = None,
    rng_seed: int | None = None,
):
  """Get IDs for brute force search, subsampling as needed."""
  ids = np.array(db.match_window_ids())
  if sample_size is None:
    return ids
  if isinstance(sample_size, float):
    if sample_size < 0 or sample_size > 1:
      raise ValueError('Float sample size must be between 0 and 1.')
    sample_size = int(sample_size * db.count_embeddings())
  np.random.default_rng(rng_seed).shuffle(ids)
  return ids[:sample_size]


def get_random_embedding_scores(
    db: interface.HopliteDBInterface,
    query_embedding: np.ndarray,
    score_fn: score_functions.ScoreFn,
    sample_size: int | float | None = 1024,
    rng_seed: int | None = None,
) -> np.ndarray:
  """Get scores for a random sample of embeddings."""
  ids = get_brute_search_ids(db, sample_size, rng_seed)
  embeddings = db.get_embeddings_batch(ids)
  scores = score_fn(embeddings, query_embedding)
  return cast(np.ndarray, scores)


def rerank(
    query_embedding: np.ndarray,
    results: search_results.TopKSearchResults,
    db: interface.HopliteDBInterface,
    score_fn: score_functions.ScoreFn,
) -> search_results.TopKSearchResults:
  """Rescore the search results using a different score function."""
  new_results = search_results.TopKSearchResults(results.top_k)
  for r in results:
    new_results.update(
        search_results.SearchResult(
            r.window_id,
            score_fn(query_embedding, db.get_embedding(r.window_id)),
        )
    )
  return new_results



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Path utilities.

General utilities to help with handling paths.
"""

import os
from typing import BinaryIO, TextIO

from etils import epath


def get_absolute_path(relative_path: os.PathLike[str] | str) -> epath.Path:
  """Returns the absolute epath.Path associated with the relative_path.

  Args:
    relative_path: The relative path (w.r.t. root) to the resource.

  Returns:
    The absolute path to the resource.
  """
  file_path = epath.Path(__file__).parent / relative_path
  return file_path


def open_file(relative_path: os.PathLike[str] | str, mode) -> TextIO | BinaryIO:
  return open(get_absolute_path(relative_path), mode)



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for colab display elements."""

import os
import shutil
import tempfile
from unittest import mock

import IPython
import ipywidgets
import numpy as np
from perch_hoplite.agile import audio_loader
from perch_hoplite.agile import embedding_display
from perch_hoplite.agile import source_info
from perch_hoplite.agile.tests import test_utils
from perch_hoplite.db import datatypes

from absl.testing import absltest


class EmbeddingDisplayTest(absltest.TestCase):

  def setUp(self):
    # Without this, unit tests using Ipywidgets will fail with 'Comms cannot be
    # opened without a kernel and a comm_manager attached to that kernel'. This
    # mocks out the comms. This is a little fragile because it sets a private
    # attribute and may break for future Ipywidget library upgrades.
    setattr(
        ipywidgets.Widget,
        '_comm_default',
        lambda self: mock.MagicMock(spec=IPython.kernel.comm.Comm),
    )

    super().setUp()
    # `self.create_tempdir()` raises an UnparsedFlagAccessError, which is why
    # we use `tempdir` directly.
    self.tempdir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  def test_query_display(self):
    classes = ['pos', 'neg']
    filenames = ['foo', 'bar', 'baz']
    sample_rate_hz = 16000
    test_utils.make_wav_files(
        self.tempdir,
        classes,
        filenames,
        file_len_s=6.0,
        sample_rate_hz=sample_rate_hz,
    )

    qd = embedding_display.QueryDisplay(
        uri=os.path.join(self.tempdir, 'pos', 'foo_pos.wav'),
        window_size_s=3.0,
        sample_rate_hz=sample_rate_hz,
    )
    # Check that we can load the audio file, and obtain the entire file.
    qd.update_audio()
    self.assertLen(qd.audio, 16000 * 6)

    qd.update_spectrogram()
    self.assertEqual(qd.full_spectrogram.shape, (600, 160))
    self.assertEqual(qd.window_spectrogram.shape, (300, 160))
    initial_spectrogram = qd.window_spectrogram.copy()

    qd.offset_s = 3.0
    qd.update_spectrogram()
    self.assertEqual(qd.window_spectrogram.shape, (300, 160))
    # Check that the spectrogram changed.
    diff = np.abs(initial_spectrogram - qd.window_spectrogram).sum()
    self.assertGreater(diff, 1.0)

    qd.offset_s = 0.0
    qd.update_spectrogram()
    self.assertEqual(qd.window_spectrogram.shape, (300, 160))
    # Check that the spectrogram is back to the original.
    diff = np.abs(initial_spectrogram - qd.window_spectrogram).sum()
    self.assertEqual(diff, 0.0)

  def test_embedding_display_button(self):
    disp = embedding_display.EmbeddingDisplay(
        window_id=123,
        recording_id=234,
        offsets=[1.0, 2.0],
        dataset_name='test_dataset',
        uri='test_uri',
        offset_s=1.0,
        score=0.5,
    )
    disp._make_label_widgets(('foo', 'bar'))
    foo_button = disp.widgets['foo']
    bar_button = disp.widgets['bar']
    self.assertIsNotNone(foo_button)
    self.assertIsNotNone(bar_button)
    # Buttons start with `None` value, indicating no label.
    # Clicking each button cycles None => POSITIVE => NEGATIVE => UNCERTAIN
    # => None.
    self.assertIsNone(foo_button.value)
    self.assertIsNone(bar_button.value)
    self.assertEmpty(disp.harvest_labels('test_provenance'))
    foo_button.click()
    self.assertEqual(foo_button.value, datatypes.LabelType.POSITIVE)
    self.assertIsNone(bar_button.value)
    bar_button.click()
    self.assertEqual(foo_button.value, datatypes.LabelType.POSITIVE)
    self.assertEqual(bar_button.value, datatypes.LabelType.POSITIVE)
    foo_button.click()
    self.assertEqual(foo_button.value, datatypes.LabelType.NEGATIVE)
    self.assertEqual(bar_button.value, datatypes.LabelType.POSITIVE)
    labels = disp.harvest_labels('test_provenance')
    self.assertLen(labels, 2)
    self.assertEqual(labels[0].recording_id, 234)
    self.assertSequenceAlmostEqual(labels[0].offsets, [1.0, 2.0])
    self.assertEqual(labels[0].label, 'foo')
    self.assertEqual(labels[0].label_type, datatypes.LabelType.NEGATIVE)
    self.assertEqual(labels[0].provenance, 'test_provenance')

    self.assertEqual(labels[1].recording_id, 234)
    self.assertSequenceAlmostEqual(labels[1].offsets, [1.0, 2.0])
    self.assertEqual(labels[1].label, 'bar')
    self.assertEqual(labels[1].label_type, datatypes.LabelType.POSITIVE)
    self.assertEqual(labels[1].provenance, 'test_provenance')

    bar_button.click()
    self.assertEqual(foo_button.value, datatypes.LabelType.NEGATIVE)
    self.assertEqual(bar_button.value, datatypes.LabelType.NEGATIVE)
    foo_button.click()
    self.assertEqual(foo_button.value, datatypes.LabelType.UNCERTAIN)
    self.assertEqual(bar_button.value, datatypes.LabelType.NEGATIVE)
    foo_button.click()
    self.assertIsNone(foo_button.value)
    self.assertEqual(bar_button.value, datatypes.LabelType.NEGATIVE)
    bar_button.click()
    self.assertIsNone(foo_button.value)
    self.assertEqual(bar_button.value, datatypes.LabelType.UNCERTAIN)
    bar_button.click()
    self.assertIsNone(foo_button.value)
    self.assertIsNone(bar_button.value)

  def test_embedding_display_group(self):
    classes = ['pos', 'neg']
    filenames = ['foo', 'bar', 'baz']
    sample_rate_hz = 16000
    test_utils.make_wav_files(
        self.tempdir,
        classes,
        filenames,
        file_len_s=6.0,
        sample_rate_hz=sample_rate_hz,
    )
    member0 = embedding_display.EmbeddingDisplay(
        window_id=123,
        recording_id=234,
        offsets=[1.0, 2.0],
        dataset_name='test_dataset',
        uri=os.path.join('pos', 'foo_pos.wav'),
        offset_s=1.0,
        score=0.5,
        sample_rate_hz=sample_rate_hz,
    )
    member1 = embedding_display.EmbeddingDisplay(
        window_id=456,
        recording_id=567,
        offsets=[2.0, 3.0],
        dataset_name='test_dataset',
        uri=os.path.join('neg', 'bar_neg.wav'),
        offset_s=2.0,
        score=0.6,
        sample_rate_hz=sample_rate_hz,
    )

    audio_source = source_info.AudioSourceConfig(
        dataset_name='test_dataset',
        base_path=self.tempdir,
        file_glob='*/*.wav',
    )
    audio_sources = source_info.AudioSources((audio_source,))
    filepath_audio_loader = audio_loader.make_filepath_loader(
        audio_sources, sample_rate_hz=sample_rate_hz, window_size_s=1.0
    )
    group = embedding_display.EmbeddingDisplayGroup.create(
        members=[member0, member1],
        sample_rate_hz=sample_rate_hz,
        audio_loader=filepath_audio_loader,
    )
    self.assertLen(group.members, 2)
    self.assertEqual(group.members[0].window_id, 123)
    self.assertEqual(group.members[1].window_id, 456)
    self.assertEqual(group.members[0].recording_id, 234)
    self.assertEqual(group.members[1].recording_id, 567)
    self.assertSequenceAlmostEqual(group.members[0].offsets, [1.0, 2.0])
    self.assertSequenceAlmostEqual(group.members[1].offsets, [2.0, 3.0])
    for got_member in group.iterator_with_audio():
      self.assertEqual(got_member.dataset_name, 'test_dataset')
      self.assertIsNotNone(got_member.audio)
      self.assertEqual(got_member.audio.shape[0], 16000)
      self.assertIsNotNone(got_member.spectrogram)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for metrics."""

import numpy as np
from perch_hoplite.agile import metrics
from absl.testing import absltest


class MetricsTest(absltest.TestCase):

  def test_average_precision_no_labels(self):
    batch_size = 4
    num_classes = 5

    np.random.seed(42)
    logits = np.random.normal(size=[batch_size, num_classes])
    labels = np.zeros_like(logits)
    av_prec = np.mean(metrics.average_precision(logits, labels))
    self.assertEqual(av_prec, 0.0)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Object wrapping an embedded audio example for display."""

from collections.abc import Callable, Iterator, Mapping, Sequence
import dataclasses
import functools

import IPython
from IPython.display import clear_output  # pylint: disable=g-importing-member
from IPython.display import display as ipy_display  # pylint: disable=g-importing-member
from IPython.display import HTML  # pylint: disable=g-importing-member
import ipywidgets
import librosa
from librosa import display as librosa_display
from matplotlib import pyplot as plt
from ml_collections import config_dict
import numpy as np
from perch_hoplite import audio_io
from perch_hoplite.db import datatypes
from perch_hoplite.db import interface
from perch_hoplite.db import search_results


@dataclasses.dataclass
class QueryDisplay:
  """Object wrapping a query for display."""

  uri: str
  window_size_s: float
  audio: np.ndarray | None = None
  full_spectrogram: np.ndarray | None = None
  window_spectrogram: np.ndarray | None = None
  offset_s: float = 0.0
  sample_rate_hz: int = 32000
  frame_rate: int = 100

  def update_audio(self):
    """Get the audio for this query."""
    if self.audio is None:
      self.audio = np.array(audio_io.load_audio(self.uri, self.sample_rate_hz))

  def update_spectrogram(self):
    """Update the spectrogram for this query."""
    if self.full_spectrogram is None:
      self.full_spectrogram = get_melspec_layer(self.sample_rate_hz)(self.audio)
    start = int(self.offset_s * self.sample_rate_hz)
    end = int(start + self.window_size_s * self.sample_rate_hz)
    self.window_spectrogram = get_melspec_layer(self.sample_rate_hz)(
        self.audio[start:end]
    )

  def get_audio_window(self) -> np.ndarray | None:
    """Get the audio window for this query."""
    if self.audio is None:
      self.update_audio()
    if self.audio is None:
      return None
    start = int(self.offset_s * self.sample_rate_hz)
    end = int(start + self.window_size_s * self.sample_rate_hz)
    return self.audio[start:end]

  def plot_spectrogram(self, spec: np.ndarray):
    fig, ax = plt.subplots()
    img = librosa_display.specshow(
        spec,
        sr=self.sample_rate_hz,
        y_axis='mel',
        x_axis='time',
        hop_length=self.sample_rate_hz // self.frame_rate,
        cmap='Greys',
        ax=ax,
    )
    return fig, img, ax

  def display(self, display_full: bool = True):
    """Display the audio, spectrogram, and label buttons."""
    self.update_audio()
    self.update_spectrogram()
    # Display full-audio spectrogram
    if display_full and self.full_spectrogram is not None:
      librosa_display.specshow(
          self.full_spectrogram.T,
          sr=self.sample_rate_hz,
          y_axis='mel',
          x_axis='time',
          hop_length=self.sample_rate_hz // self.frame_rate,
          cmap='Greys',
      )
      window_end = self.offset_s + self.window_size_s
      plt.plot([self.offset_s, self.offset_s], [0, 16000], 'r:')
      plt.plot([window_end, window_end], [0, 16000], 'r:')
      plt.show()

    # Display the current window.
    if self.window_spectrogram is not None:
      librosa_display.specshow(
          self.window_spectrogram.T,
          sr=self.sample_rate_hz,
          y_axis='mel',
          x_axis='time',
          hop_length=self.sample_rate_hz // self.frame_rate,
          cmap='Greys',
      )
      plt.show()

    if display_full:
      ipy_display(IPython.display.Audio(self.audio, rate=self.sample_rate_hz))
    window_st = int(self.offset_s * self.sample_rate_hz)
    window_end = int(window_st + self.window_size_s * self.sample_rate_hz)
    audio_window = self.audio[window_st:window_end]
    ipy_display(IPython.display.Audio(audio_window, rate=self.sample_rate_hz))

  def display_interactive(self):
    """Create an interactive slider for the offset."""
    self.update_audio()
    if self.audio is None:
      print(f'No audio loaded for {self.uri}.')
      return
    self.update_spectrogram()
    slider = ipywidgets.FloatSlider(
        value=self.offset_s,
        min=0.0,
        max=self.audio.shape[0] / self.sample_rate_hz,
        continuous_update=False,
        description='offset_s',
    )

    def update(x):
      self.offset_s = x
      clear_output(wait=True)
      self.display()

    ipywidgets.interact(update, x=slider)


@dataclasses.dataclass
class EmbeddingDisplay:
  """Object wrapping an embedded audio example for display."""

  window_id: int
  recording_id: int
  offsets: list[float]
  dataset_name: str
  uri: str
  offset_s: float
  score: float
  widgets: dict[str, ipywidgets.Button] = dataclasses.field(
      default_factory=dict
  )
  sample_rate_hz: int = 32000
  frame_rate: int = 100
  audio: np.ndarray | None = None
  spectrogram: np.ndarray | None = None
  labels: Sequence[datatypes.Annotation] = ()
  init_label_type: datatypes.LabelType | None = None

  LABEL_TYPE_TO_BUTTON_INFO = {
      None: dict(icon='', button_style=''),
      datatypes.LabelType.POSITIVE: dict(
          icon='plus-circle', button_style='success'
      ),
      datatypes.LabelType.NEGATIVE: dict(
          icon='minus-circle', button_style='warning'
      ),
      datatypes.LabelType.UNCERTAIN: dict(
          icon='question-circle', button_style='info'
      ),
  }

  def _make_label_button(self, button_label: str) -> ipywidgets.Button:
    """Create an ipywidget button for the given label."""

    def button_callback(x):
      if x.value is None:  # uninitialized => positive
        x.value = datatypes.LabelType.POSITIVE
        x.icon = 'plus-circle'
        x.button_style = 'success'
      elif x.value == datatypes.LabelType.POSITIVE:  # positive => negative
        x.value = datatypes.LabelType.NEGATIVE
        x.icon = 'minus-circle'
        x.button_style = 'warning'
      elif x.value == datatypes.LabelType.NEGATIVE:  # negative => uncertain
        x.value = datatypes.LabelType.UNCERTAIN
        x.icon = 'question-circle'
        x.button_style = 'info'
      elif (
          x.value == datatypes.LabelType.UNCERTAIN
      ):  # uncertain => uninitialized
        x.value = None
        x.icon = ''
        x.button_style = ''
      else:
        raise ValueError(f'Unexpected button value: {x.value}')

    button_info = self.LABEL_TYPE_TO_BUTTON_INFO[self.init_label_type]
    button = ipywidgets.Button(
        description=button_label,
        icon=button_info['icon'],
        button_style=button_info['button_style'],
    )
    button.value = self.init_label_type
    button.on_click(button_callback)
    return button

  def _make_label_widgets(self, labels: Sequence[str]):
    """Create ipywidget buttons for the given labels."""
    # Create widgets for the labels as needed.
    for label in labels:
      if label not in self.widgets:
        self.widgets[label] = self._make_label_button(label)

  def display(
      self,
      labels: Sequence[str] = (),
      rank: int = -1,
      show_score: bool = True,
      button_columns: int = 4,
      show_labels: bool = False,
  ):
    """Display the audio, spectrogram, and label buttons."""
    self._make_label_widgets(labels)
    if self.audio is None:
      print(f'No audio loaded for {self.uri} :: {self.offset_s}.')
      return
    if self.spectrogram is None:
      print(f'No spectrogram computed for {self.uri} :: {self.offset_s}.')
      return
    # Display spectrogram
    librosa_display.specshow(
        self.spectrogram.T,
        sr=self.sample_rate_hz,
        y_axis='mel',
        x_axis='time',
        hop_length=self.sample_rate_hz // self.frame_rate,
        cmap='Greys',
    )
    plt.show()
    # Display audio
    ipy_display(IPython.display.Audio(self.audio, rate=self.sample_rate_hz))
    print(f'dataset name : {self.dataset_name}')
    print(f'source uri   : {self.uri}')
    print(f'offset_s     : {self.offset_s:.2f}')
    if rank > 0:
      print(f'rank         : {rank}')
    if show_score:
      print(f'score        : {(self.score):.2f}')
    if show_labels:
      for lbl in self.labels:
        print(f'\t{lbl.label}: {lbl.label_type.value}')

    # Display widgets
    grid = ipywidgets.GridspecLayout(
        n_rows=len(self.widgets) // button_columns + 1,
        n_columns=button_columns,
    )
    for i, button in enumerate(self.widgets.values()):
      if button is None:
        continue
      row = i // button_columns
      col = i % button_columns
      grid[row, col] = button
    ipy_display(
        HTML(
            '<link rel="stylesheet"'
            ' href="//stackpath.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css"/>'
        )
    )
    ipy_display(grid)

  def harvest_labels(
      self,
      provenance: str,
      skip_uncertain: bool = True,
  ) -> Sequence[datatypes.Annotation]:
    """Get the labels for this example."""

    labels = []
    for lbl, w in self.widgets.items():
      if w.value is None:  # uninitialized
        continue
      if w.value == datatypes.LabelType.UNCERTAIN and skip_uncertain:
        continue
      labels.append(
          datatypes.Annotation(
              id=-1,
              recording_id=self.recording_id,
              offsets=self.offsets,
              label=lbl,
              label_type=w.value,
              provenance=provenance,
          )
      )
    return labels

  def harvest_annotated_window(
      self,
      provenance: str,
      skip_uncertain: bool = True,
  ) -> tuple[int, Sequence[datatypes.Annotation]]:
    """Get the annotated windows for this example."""

    labels = []
    for lbl, w in self.widgets.items():
      if w.value is None:  # uninitialized
        continue
      if w.value == datatypes.LabelType.UNCERTAIN and skip_uncertain:
        continue
      labels.append(
          datatypes.Annotation(
              id=-1,
              recording_id=self.recording_id,
              offsets=self.offsets,
              label=lbl,
              label_type=w.value,
              provenance=provenance,
          )
      )
    return self.window_id, labels


@dataclasses.dataclass
class EmbeddingDisplayGroup:
  """Group of EmbeddingDisplay objects."""

  members: Sequence[EmbeddingDisplay]
  melspec_layer: Callable[[np.ndarray], np.ndarray]
  audio_loader: Callable[[str, float], np.ndarray]
  current_page: int = 0
  results_per_page: int = 10
  sample_rate_hz: int = 32000

  @classmethod
  def create(
      cls, members: Sequence[EmbeddingDisplay], sample_rate_hz: int, **kwargs
  ) -> 'EmbeddingDisplayGroup':
    """Create an EmbeddingDisplayGroup from a list of EmbeddingDisplay objects."""
    melspec_layer = get_melspec_layer(sample_rate_hz)
    return cls(
        members=members,
        melspec_layer=melspec_layer,
        sample_rate_hz=sample_rate_hz,
        **kwargs,
    )

  @classmethod
  def from_search_results(
      cls,
      results: search_results.TopKSearchResults,
      db: interface.HopliteDBInterface,
      sample_rate_hz: int,
      frame_rate: int,
      filter_labels: Sequence[str] = (),
      **kwargs,
  ) -> 'EmbeddingDisplayGroup':
    """Create an EmbeddingDisplayGroup from a Hoplite TopKSearchResults object."""
    members = []
    for result in results:
      window_id = int(result.window_id)
      window = db.get_window(window_id)
      recording = db.get_recording(window.recording_id)
      deployment = db.get_deployment(recording.deployment_id)
      annotations = db.get_all_annotations(
          filter=config_dict.create(
              eq=dict(recording_id=window.recording_id),
              approx=dict(offsets=window.offsets),
          )
      )
      if [a for a in annotations if a.label in filter_labels]:
        continue
      members.append(
          EmbeddingDisplay(
              window_id=window_id,
              recording_id=window.recording_id,
              offsets=window.offsets,
              dataset_name=deployment.project,
              uri=recording.filename,
              offset_s=window.offsets[0],
              score=result.sort_score,
              sample_rate_hz=sample_rate_hz,
              frame_rate=frame_rate,
              labels=annotations,
              init_label_type=result.display_label_type,
          )
      )
    return cls.create(members=members, sample_rate_hz=sample_rate_hz, **kwargs)

  def iterator_with_audio(
      self,
      current_page_only: bool = False,
  ) -> Iterator[EmbeddingDisplay]:
    """Fetch audio for members of the group and compute spectrograms."""
    if len(self.members) > self.results_per_page and current_page_only:
      min_idx = self.current_page * self.results_per_page
      max_idx = min(
          (self.current_page + 1) * self.results_per_page, len(self.members)
      )
      targets = self.members[min_idx:max_idx]
    else:
      targets = self.members

    # We want to iterate over all members, but only load audio for those that
    # don't already have it.
    needs_audio = [m.audio is None for m in targets]
    needs_audio_targets = [m for m in targets if m.audio is None]

    filepaths = [m.uri for m in needs_audio_targets]
    offsets = [m.offset_s for m in needs_audio_targets]
    audio_iter_ = audio_io.multi_load_audio_window(
        filepaths, offsets, self.audio_loader
    )
    for member, is_dispatched in zip(targets, needs_audio):
      if is_dispatched:
        got_audio = next(audio_iter_)
        if got_audio is not None:
          member.audio = got_audio
          member.spectrogram = self.melspec_layer(got_audio)
      yield member

  @property
  def num_pages(self) -> int:
    return len(self.members) // self.results_per_page

  def increment_page(self, inc: int):
    self.current_page += inc
    self.current_page = min(self.current_page, self.num_pages)
    self.current_page = max(0, self.current_page)

  def display(
      self,
      positive_labels: Sequence[str] = (),
      show_score: bool = True,
      paged_mode: bool = True,
      show_labels: bool = False,
  ):
    """Display the audio, spectrogram, and label buttons."""
    clear_output()
    if paged_mode:
      rank_offset = self.current_page * self.results_per_page
      print(f'Page {self.current_page + 1} of {self.num_pages + 1}')
      print('-' * 80 + '\n')
    else:
      rank_offset = 0

    # Attach audio as needed.
    member_iterator = self.iterator_with_audio(current_page_only=paged_mode)

    # Display the selected members.
    for r, member in enumerate(member_iterator):
      member.display(
          positive_labels,
          show_score=show_score,
          rank=r + rank_offset,
          button_columns=1,
          show_labels=show_labels,
      )
      print('\n' + '-' * 80)

    # Display next/prev page buttons as needed.
    if paged_mode:
      print(f'Page {self.current_page + 1} of {self.num_pages + 1}')

      def increment_page_callback(unused_x, inc):
        self.increment_page(inc)
        self.display(
            positive_labels, show_score=show_score, paged_mode=paged_mode
        )

      next_page_button = ipywidgets.Button(description='Next Page')
      next_page_button.on_click(lambda x: increment_page_callback(x, 1))
      prev_page_button = ipywidgets.Button(description='Prev Page')
      prev_page_button.on_click(lambda x: increment_page_callback(x, -1))
      ipy_display(prev_page_button)
      ipy_display(next_page_button)

  def harvest_labels(
      self,
      provenance: str,
      skip_uncertain: bool = True,
  ) -> Sequence[datatypes.Annotation]:
    labels = []
    for member in self.members:
      labels.extend(member.harvest_labels(provenance, skip_uncertain))
    return labels

  def harvest_annotated_windows(
      self,
      provenance: str,
      skip_uncertain: bool = True,
  ) -> Mapping[int, Sequence[datatypes.Annotation]]:
    annotations_dict = {}
    for member in self.members:
      window_id, annotations = member.harvest_annotated_window(
          provenance, skip_uncertain
      )
      annotations_dict[window_id] = annotations
    return annotations_dict


def pcen_melspec_display(
    audio: np.ndarray, sample_rate_hz: int, root: float = 4.0
):
  """Compute a PCEN mel spectrogram for display."""
  stride = sample_rate_hz // 100
  spec = librosa.feature.melspectrogram(
      y=audio,
      sr=sample_rate_hz,
      hop_length=stride,
      win_length=4 * stride,
      center=True,
      n_mels=160,
      power=1,
  )
  pspec = librosa.pcen(
      spec,
      sr=sample_rate_hz,
      hop_length=stride,
      gain=0.8,
      bias=0.01,
      power=1.0 / root,
      axis=0,
  )
  return pspec.T[:-1]


@functools.cache
def get_melspec_layer(
    sample_rate_hz: int, root=4.0
) -> Callable[[np.ndarray], np.ndarray]:
  """Creates a melspec layer for easy visualization."""
  return lambda x: pcen_melspec_display(
      x, sample_rate_hz=sample_rate_hz, root=root
  )



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Audio loader function helpers."""

import os
from typing import Callable

from etils import epath
import numpy as np
from perch_hoplite import audio_io
from perch_hoplite.agile import source_info


def make_filepath_loader(
    audio_sources: source_info.AudioSources,
    sample_rate_hz: int = 32000,
    window_size_s: float = 5.0,
    dtype: str = 'float32',
) -> Callable[[str, float], np.ndarray]:
  """Create a function for loading audio from a source ID and offset.

  Note that if multiple globs match a given source ID, the first match is used.

  Args:
    audio_sources: Embedding audio sources.
    sample_rate_hz: Sample rate of the audio.
    window_size_s: Window size of the audio.
    dtype: Data type of the audio.

  Returns:
    Function for loading audio from a source ID and offset.

  Raises:
    ValueError if no audio path is found for the given source ID.
  """

  def loader(source_id: str, offset_s: float) -> np.ndarray:
    found_path = None
    for audio_source in audio_sources.audio_globs:
      path = epath.Path(audio_source.base_path) / source_id
      if path.exists():
        found_path = path
        break
    if found_path is None:
      raise ValueError('No audio path found for source_id: ', source_id)
    return np.array(
        audio_io.load_audio_window(
            found_path.as_posix(),
            offset_s,
            sample_rate=sample_rate_hz,
            window_size_s=window_size_s,
        ),
        dtype=dtype,
    )

  return loader



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convenience utilities for handling class lists."""

from perch_hoplite.taxonomy import namespace_db


def get_class_lists(species_class_list_name: str, add_taxonomic_labels: bool):
  """Get the number of classes for the target class outputs."""
  db = namespace_db.load_db()
  species_classes = db.class_lists[species_class_list_name]
  class_lists = {
      "label": species_classes,
  }
  if add_taxonomic_labels:
    for name in ["genus", "family", "order"]:
      mapping_name = f"{species_classes.namespace}_to_{name}"
      mapping = db.mappings[mapping_name]
      taxa_class_list = species_classes.apply_namespace_mapping(mapping)
      class_lists[name] = taxa_class_list
  return class_lists



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Hoplite databases."""

import shutil
import tempfile

from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import brutalism
from perch_hoplite.db import datatypes
from perch_hoplite.db import iterators
from perch_hoplite.db.tests import test_utils

from absl.testing import absltest
from absl.testing import parameterized

EMBEDDING_SIZE = 128


class HopliteTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.tempdir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  @parameterized.product(
      db_type=test_utils.DB_TYPES,
      thread_split=(True, False),
  )
  def test_graph_db_interface(self, db_type, thread_split):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, db_type, 1000, rng, EMBEDDING_SIZE)
    if thread_split:
      db = db.thread_split()

    # Run all methods in the interface...
    self.assertEqual(db.count_embeddings(), 1000)
    idxes = db.match_window_ids()

    one_idx = db.match_window_ids(limit=1)[0]
    self.assertIn(one_idx, idxes)

    # Check the metadata.
    got_md = db.get_metadata('db_config')
    self.assertEqual(got_md.embedding_dim, EMBEDDING_SIZE)

    with self.subTest('test_embedding_sources'):
      recording_id = db.get_window(window_id=idxes[1]).recording_id
      recording = db.get_recording(recording_id)
      deployment = db.get_deployment(recording.deployment_id)
      # The embeddings are given one of three randomly selected dataset names.
      embs = db.match_window_ids(
          deployments_filter=config_dict.create(
              eq=dict(project=deployment.project)
          ),
      )
      self.assertGreater(len(embs), db.count_embeddings() / 6)
      # For an unknown dataset name, we should get no embeddings.
      embs = db.match_window_ids(
          deployments_filter=config_dict.create(eq=dict(project='fake_name'))
      )
      self.assertEmpty(embs)
      # Source ids are approximately unique.
      embs = db.match_window_ids(
          deployments_filter=config_dict.create(
              eq=dict(project=deployment.project)
          ),
          recordings_filter=config_dict.create(
              eq=dict(filename=recording.filename)
          ),
      )
      self.assertLen(embs, 1)
      # For an unknown source id, we should get no embeddings.
      embs = db.match_window_ids(
          deployments_filter=config_dict.create(
              eq=dict(project=deployment.project)
          ),
          recordings_filter=config_dict.create(eq=dict(filename='fake_id')),
      )
      self.assertEmpty(embs)
    db.commit()

  @parameterized.product(db_type=test_utils.PERSISTENT_DB_TYPES)
  def test_persistence(self, db_type):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, db_type, 1000, rng, EMBEDDING_SIZE)
    one_emb = np.random.normal(size=(EMBEDDING_SIZE,), loc=0, scale=0.05)
    one_dep_id = db.insert_deployment(name='q', project='q')
    one_rec_id = db.insert_recording(filename='x', deployment_id=one_dep_id)
    one_emb_id = db.insert_window(
        recording_id=one_rec_id, offsets=[5.0], embedding=one_emb
    )
    self.assertLen(db.match_window_ids(), 1001)
    db.commit()

    got_emb = db.get_embedding(one_emb_id)
    np.testing.assert_equal(got_emb, np.float16(one_emb))

    # "Making" the persistent DB without adding any new embeddings gives us a
    # view of the saved DB.
    test_db = test_utils.make_db(self.tempdir, db_type, 0, rng, EMBEDDING_SIZE)
    self.assertIn(one_emb_id, test_db.match_window_ids())
    # Check that the embeddings are the same in the two DB's.
    for idx in db.match_window_ids():
      emb = db.get_embedding(idx)
      test_emb = test_db.get_embedding(idx)
      np.testing.assert_equal(emb, test_emb)

  @parameterized.product(db_type=test_utils.DB_TYPES)
  def test_labels_db_interface(self, db_type):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, db_type, 1000, rng, EMBEDDING_SIZE)
    windows = db.get_all_windows()
    db.insert_annotation(
        recording_id=windows[0].recording_id,
        offsets=windows[0].offsets,
        label='hawgoo',
        label_type=datatypes.LabelType.POSITIVE,
        provenance='human',
    )
    db.insert_annotation(
        recording_id=windows[0].recording_id,
        offsets=windows[0].offsets,
        label='hawgoo',
        label_type=datatypes.LabelType.POSITIVE,
        provenance='machine',
    )
    db.insert_annotation(
        recording_id=windows[1].recording_id,
        offsets=windows[1].offsets,
        label='hawgoo',
        label_type=datatypes.LabelType.POSITIVE,
        provenance='machine',
    )
    db.insert_annotation(
        recording_id=windows[0].recording_id,
        offsets=windows[0].offsets,
        label='rewbla',
        label_type=datatypes.LabelType.NEGATIVE,
        provenance='machine',
    )

    with self.subTest('get_embeddings_by_label'):
      # When both label_type and provenance are unspecified, we should get all
      # unique IDs with the target label. Id's 0 and 1 both have some kind of
      # 'hawgoo' label.
      got = db.match_window_ids(
          annotations_filter=config_dict.create(eq=dict(label='hawgoo'))
      )
      self.assertSequenceEqual(
          sorted(got), sorted([windows[0].id, windows[1].id])
      )

    with self.subTest('get_embeddings_by_label_type'):
      # Now we should get the ID's for all POSITIVE 'hawgoo' labels, regardless
      # of provenance.
      got = db.match_window_ids(
          annotations_filter=config_dict.create(
              eq=dict(label='hawgoo', label_type=datatypes.LabelType.POSITIVE)
          ),
      )
      self.assertSequenceEqual(
          sorted(got), sorted([windows[0].id, windows[1].id])
      )

      # There are no negative 'hawgoo' labels.
      got = db.match_window_ids(
          annotations_filter=config_dict.create(
              eq=dict(label='hawgoo', label_type=datatypes.LabelType.NEGATIVE)
          )
      )
      self.assertEmpty(got)

    with self.subTest('get_embeddings_by_label_provenance'):
      # There is only one hawgoo labeled by a human.
      got = db.match_window_ids(
          annotations_filter=config_dict.create(
              eq=dict(label='hawgoo', provenance='human')
          )
      )
      self.assertSequenceEqual(got, [windows[0].id])

      # And only one example with a 'rewbla' labeled by a machine.
      got = db.match_window_ids(
          annotations_filter=config_dict.create(
              eq=dict(label='rewbla', provenance='machine')
          )
      )
      self.assertSequenceEqual(got, [windows[0].id])

    with self.subTest('count_all_labels'):
      # Finally, there are a total of three labels on ID 0.
      got = db.get_all_annotations(
          config_dict.create(eq=dict(recording_id=windows[0].recording_id))
      )
      self.assertLen(got, 3)

    with self.subTest('get_classes'):
      got = db.get_all_labels()
      self.assertSequenceEqual(got, ['hawgoo', 'rewbla'])

    with self.subTest('get_class_counts'):
      # 2 positive labels for 'hawgoo' ignoring provenance.
      got = db.count_each_label(datatypes.LabelType.POSITIVE)
      self.assertDictEqual(got, {'hawgoo': 2})

      # 1 negative label for 'rewbla'.
      got = db.count_each_label(datatypes.LabelType.NEGATIVE)
      self.assertDictEqual(got, {'rewbla': 1})

      # 2 labels for 'hawgoo' and 1 for 'rewbla' when counting all labels,
      # regardless of type.
      got = db.count_each_label()
      self.assertDictEqual(got, {'hawgoo': 2, 'rewbla': 1})

    with self.subTest('count_classes'):
      self.assertLen(db.get_all_labels(), 2)

    with self.subTest('duplicate_labels'):
      annotation_id = db.insert_annotation(
          recording_id=windows[0].recording_id,
          offsets=windows[0].offsets,
          label='unique',
          label_type=datatypes.LabelType.POSITIVE,
          provenance='human',
      )
      dupe_annotation_id = db.insert_annotation(
          recording_id=windows[0].recording_id,
          offsets=windows[0].offsets,
          label='unique',
          label_type=datatypes.LabelType.POSITIVE,
          provenance='human',
          handle_duplicates='skip',
      )
      self.assertEqual(annotation_id, dupe_annotation_id)

  @parameterized.named_parameters(*test_utils.DB_TYPE_NAMED_PAIRS)
  def test_brute_search_impl_agreement(self, target_db_type, source_db_type):
    rng = np.random.default_rng(42)
    source_db = test_utils.make_db(
        self.tempdir, source_db_type, 1000, rng, EMBEDDING_SIZE
    )
    target_db = test_utils.make_db(
        self.tempdir,
        target_db_type,
        1000,
        rng,
        EMBEDDING_SIZE,
        fill_random=False,
    )
    id_mapping = test_utils.clone_embeddings(source_db, target_db)

    # Check brute-force search agreement.
    query = rng.normal(size=(128,), loc=0, scale=1.0)
    results_m = brutalism.brute_search(
        source_db, query, search_list_size=10, score_fn=np.dot
    )
    results_s = brutalism.brute_search(
        target_db, query, search_list_size=10, score_fn=np.dot
    )
    self.assertLen(results_m.search_results, 10)
    self.assertLen(results_s.search_results, 10)
    # Search results are iterated over in sorted order.
    for r_m, r_s in zip(results_m, results_s):
      emb_m = source_db.get_embedding(r_m.window_id)
      emb_s = target_db.get_embedding(r_s.window_id)
      self.assertEqual(id_mapping[r_m.window_id], r_s.window_id)
      # TODO(tomdenton): check that the scores are the same.
      np.testing.assert_equal(emb_m, emb_s)

  def test_random_batched_iterator(self):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, 'in_mem', 100, rng, EMBEDDING_SIZE)
    # Test behavior when batch size divides the number of embeddings.
    with self.subTest('batch_size_divides_num_embeddings'):
      batch_size = 10
      iterator = iterators.random_batched_iterator(
          db.match_window_ids(), batch_size=batch_size, rng=rng
      )
      all_ids = set()
      for i in range(10):
        batch = next(iterator)
        self.assertLen(batch, batch_size)
        self.assertContainsSubset(batch, db.match_window_ids())
        all_ids.update(batch)
        self.assertLen(all_ids, min(100, (i + 1) * batch_size))

    # Test behavior when batch size does not divide the number of embeddings.
    with self.subTest('batch_size_does_not_divide_num_embeddings'):
      batch_size = 27
      iterator = iterators.random_batched_iterator(
          db.match_window_ids(), batch_size=batch_size, rng=rng
      )
      all_ids = set()
      for i in range(4):
        batch = next(iterator)
        self.assertLen(batch, batch_size)
        all_ids.update(batch)
        self.assertContainsSubset(batch, db.match_window_ids())
        self.assertLen(all_ids, min(100, (i + 1) * batch_size))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for conversion from previous agile modeling format.

Generating the embeddings requires a call to the Chirp inference library,
which creates a dependency that we don't want to include in the package
dependencies. Instead, we generate the embeddings once and check them into
the repo as a testdata file.
"""

import os
import shutil
import tempfile

from etils import epath
from ml_collections import config_dict
from perch_hoplite import audio_io
from perch_hoplite import path_utils
from perch_hoplite.agile.migrations import convert_legacy
from perch_hoplite.agile.tests import test_utils

from absl.testing import absltest
from absl.testing import parameterized


class ConvertLegacyTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    # `self.create_tempdir()` raises an UnparsedFlagAccessError, which is why
    # we use `tempdir` directly.
    self.tempdir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  def write_placeholder_embeddings(
      self, audio_glob, classes, filenames, embed_dir
  ):
    """Method used to generate embeddings for testing."""
    from chirp.inference import embed_lib
    from chirp.inference import tf_examples

    source_infos = embed_lib.create_source_infos([audio_glob], shard_len_s=-1)
    self.assertLen(source_infos, len(classes) * len(filenames))

    # Set up embedding function.
    config = config_dict.ConfigDict()
    model_kwargs = {
        'sample_rate': 16000,
        'embedding_size': 128,
        'make_embeddings': True,
        'make_logits': False,
        'make_separated_audio': False,
        'window_size_s': 5.0,
        'hop_size_s': 5.0,
    }
    embed_fn_config = config_dict.ConfigDict()
    embed_fn_config.write_embeddings = True
    embed_fn_config.write_logits = False
    embed_fn_config.write_separated_audio = False
    embed_fn_config.write_raw_audio = False
    embed_fn_config.write_frontend = False
    embed_fn_config.model_key = 'placeholder_model'
    embed_fn_config.model_config = model_kwargs
    embed_fn_config.min_audio_s = 0.1
    embed_fn_config.file_id_depth = 1
    config.embed_fn_config = embed_fn_config
    config.source_file_patterns = [audio_glob]
    config.num_shards_per_file = -1
    config.shard_len_s = -1

    epath.Path(embed_dir).mkdir(parents=True, exist_ok=True)
    embed_lib.maybe_write_config(config, epath.Path(embed_dir))

    embed_fn = embed_lib.EmbedFn(**embed_fn_config)
    embed_fn.setup()

    # Write embeddings.
    audio_loader = lambda fp, offset: audio_io.load_audio_window(
        fp, offset, sample_rate=16000, window_size_s=120.0
    )
    audio_iterator = audio_io.multi_load_audio_window(
        filepaths=[s.filepath for s in source_infos],
        offsets=[0 for _ in source_infos],
        audio_loader=audio_loader,
    )
    with tf_examples.EmbeddingsTFRecordMultiWriter(
        output_dir=embed_dir, num_files=1
    ) as file_writer:
      for source_info, audio in zip(source_infos, audio_iterator):
        file_id = source_info.file_id(1)
        offset_s = source_info.shard_num * source_info.shard_len_s
        example = embed_fn.audio_to_example(file_id, offset_s, audio)
        file_writer.write(example.SerializeToString())
      file_writer.flush()

  @parameterized.product(
      db_type=(
          'in_mem',
          'sqlite_usearch',
      ),
      generate_embeddings=(False,),
  )
  def test_convert_legacy(self, db_type, generate_embeddings):
    classes = ['pos', 'neg']
    filenames = ['foo', 'bar', 'baz']
    audio_glob = test_utils.make_wav_files(
        self.tempdir, classes, filenames, file_len_s=60.0
    )
    if generate_embeddings:
      # TODO(tomdenton): Compare the checked-in embeddings to the generated
      # embeddings to ensure that they are the same.
      embed_dir = os.path.join(self.tempdir, 'embeddings')
      self.write_placeholder_embeddings(
          audio_glob, classes, filenames, embed_dir
      )
    else:
      config_path = path_utils.get_absolute_path(
          'agile/tests/testdata/config.json'
      )
      embed_dir = epath.Path(config_path).parent

    if db_type == 'sqlite_usearch':
      kwargs = {'db_path': self.tempdir}
    elif db_type == 'in_mem':
      kwargs = {
          'max_size': 100,
          'degree_bound': 10,
      }
    else:
      raise ValueError(f'Unknown db type: {db_type}')

    db = convert_legacy.convert_tfrecords(
        embeddings_path=embed_dir,
        db_type=db_type,
        dataset_name='test',
        **kwargs,
    )
    # There are six one-minute test files, so we should get 72 embeddings.
    self.assertEqual(db.count_embeddings(), 72)
    metadata = db.get_metadata(key=None)
    self.assertIn('legacy_config', metadata)
    self.assertIn('audio_sources', metadata)
    self.assertIn('model_config', metadata)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tooling for measuring call density."""

import collections
import dataclasses
from typing import Sequence

from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import datatypes
from perch_hoplite.db import search_results
import scipy


@dataclasses.dataclass
class ValidationExample:
  """Wrapper for validation data used for call density estimation.

  Attributes:
    window_id: Window ID corresponding to the example.
    label_type: Label type (positive, negative or uncertain). Can be None if the
      example is unlabeled.
    score: Classifier score.
    bin: Bin number the validation example was assigned to.
    bin_weight: Proportion of total data in the same bin as this example.
  """

  window_id: int
  label_type: datatypes.LabelType | None
  score: float
  bin: int
  bin_weight: float


@dataclasses.dataclass
class CallDensityConfig(datatypes.HopliteConfig):
  """Config dataclass for call density estimation."""

  study_name: str
  classifier: config_dict.ConfigDict
  target_class: str
  quantile_bounds: list[float]
  samples_per_bin: int

  window_ids: dataclasses.InitVar[list[int]]
  logits: dataclasses.InitVar[list[float]]
  annotations: dataclasses.InitVar[list[datatypes.Annotation | None] | None] = (
      None
  )

  # TODO(stefanistrate): Allow setting the value bounds directly (instead of
  # computing them from the quantile bounds).
  value_bounds: list[float] = dataclasses.field(init=False)
  bins_dict: dict[int, list[ValidationExample]] = dataclasses.field(init=False)

  deployments_filter: config_dict.ConfigDict | None = None
  recordings_filter: config_dict.ConfigDict | None = None
  windows_filter: config_dict.ConfigDict | None = None
  annotations_filter: config_dict.ConfigDict | None = None

  def __post_init__(
      self,
      window_ids: list[int],
      logits: list[float],
      annotations: list[datatypes.Annotation | None] | None,
  ) -> None:
    """Post-init initialization."""

    if annotations is None:
      annotations = [None] * len(window_ids)

    self.value_bounds = np.quantile(logits, self.quantile_bounds).tolist()

    bin_numbers = np.digitize(logits, self.value_bounds)
    bin_weights = np.diff(self.quantile_bounds, prepend=-1.0)

    self.bins_dict = {bin: [] for bin in range(1, len(self.value_bounds))}
    for bin_number, window_id, logit, annotation in zip(
        bin_numbers, window_ids, logits, annotations
    ):
      if bin_number == 0:
        bin_number = 1
      elif bin_number == len(self.value_bounds):
        bin_number -= 1
      self.bins_dict[bin_number].append(
          ValidationExample(
              window_id=int(window_id),
              label_type=annotation.label_type if annotation else None,
              score=float(logit),
              bin=int(bin_number),
              bin_weight=float(bin_weights[bin_number]),
          )
      )

  def convert_bins_to_search_results(
      self,
      samples_per_bin: int | None = None,
      skip_annotated: bool = False,
  ) -> search_results.TopKSearchResults:
    """Convert bins to search results."""

    if samples_per_bin is None:
      samples_per_bin = self.samples_per_bin

    results = []
    for examples in self.bins_dict.values():
      for ex in examples[:samples_per_bin]:
        if skip_annotated and ex.label_type is not None:
          continue
        results.append(
            search_results.SearchResult(
                window_id=ex.window_id,
                sort_score=ex.score,
                display_label_type=ex.label_type,
            )
        )

    return search_results.TopKSearchResults(
        top_k=len(results), search_results=results
    )

  def select_validation_examples(
      self,
      label_types: Sequence[datatypes.LabelType | None] = (
          datatypes.LabelType.POSITIVE,
          datatypes.LabelType.NEGATIVE,
      ),
  ) -> list[ValidationExample]:
    """Select validation examples with the given label types.

    Args:
      label_types: Label types to select from the study. Label types can be one
        of the LabelType enum values (POSITIVE, NEGATIVE, UNCERTAIN), or None
        (in which case unlabeled examples are included as well). By default,
        only POSITIVE and NEGATIVE examples are selected.

    Returns:
      List of ValidationExample objects.
    """

    validation_examples = []
    for examples in self.bins_dict.values():
      for ex in examples:
        if ex.label_type in label_types:
          validation_examples.append(ex)
    return validation_examples

  def update_from_annotated_windows(
      self,
      annotations_dict: dict[int, list[datatypes.Annotation]],
  ) -> None:
    """Update examples from annotations."""

    for examples in self.bins_dict.values():
      for ex in examples:
        annotations = annotations_dict.get(ex.window_id)
        if annotations:
          ex.label_type = annotations[0].label_type


def estimate_call_density(
    examples: list[ValidationExample],
    num_beta_samples: int = 10_000,
    beta_prior: float = 0.1,
) -> tuple[float, np.ndarray]:
  """Estimates call density from a set of ValidationExample.

  Args:
    examples: Validated examples.
    num_beta_samples: Number of times to draw from beta distributions.
    beta_prior: Prior for beta distribution.

  Returns:
    Expected value of density and an array of all sampled density estimates.
  """

  # Collect validated labels by bin.
  bin_pos = collections.defaultdict(int)
  bin_neg = collections.defaultdict(int)
  bin_weights = collections.defaultdict(float)
  for ex in examples:
    bin_weights[ex.bin] = ex.bin_weight
    if ex.label_type == datatypes.LabelType.POSITIVE:
      bin_pos[ex.bin] += 1
    elif ex.label_type == datatypes.LabelType.NEGATIVE:
      bin_neg[ex.bin] += 1

  # Create beta distributions.
  betas = {}
  for b in bin_weights:
    betas[b] = scipy.stats.beta(
        bin_pos[b] + beta_prior, bin_neg[b] + beta_prior
    )

  # MLE positive rate in each bin.
  density_ev = np.array([
      bin_weights[b] * bin_pos[b] / (bin_pos[b] + bin_neg[b] + 1e-6)
      for b in bin_weights
  ]).sum()

  q_betas = []
  for _ in range(num_beta_samples):
    q_beta = np.array([
        bin_weights[b] * betas[b].rvs(size=1)[0] for b in betas  # p(b) * P(+|b)
    ]).sum()
    q_betas.append(q_beta)

  return density_ev, np.array(q_betas)


def estimate_roc_auc(
    examples: list[ValidationExample],
) -> float:
  """Estimate the classifier ROC-AUC from validation logs.

  We use the probabilistic interpretation of ROC-AUC, as the probability that
  a uniformly sampled positive example has higher score than a uniformly sampled
  negative example.

  Abusing notation a bit, we decompose this as follows:
  P(+ > -) = sum_{b,c} P(+ > - | + in b, - in c) * P(b|+) * P(c|-)
  using the law of total probability and independence of drawing the pos/neg
  examples uniformly at random.

  When comparing scores from different bins b > c, we have all scores from b
  greater than scores from c, so P(+ > - | + in b, - in c) = 1.
  Likewise, if b < c, P(+ > - | + in b, - in c) = 0.

  When b == c, we can count directly the number of + > - pairs to estimate the
  in-bin ROC-AUC.

  The scalars P(b|+) and P(c|-) are computed using Bayes' rule and the expected
  value estimate of P(+).

  Args:
    examples: List of ValidationExample objects.

  Returns:
    ROC-AUC estimate.
  """

  # Collect validated labels by bin.
  bin_pos = collections.defaultdict(int)
  bin_neg = collections.defaultdict(int)
  bin_weights = collections.defaultdict(float)
  for ex in examples:
    bin_weights[ex.bin] = ex.bin_weight
    if ex.label_type == datatypes.LabelType.POSITIVE:
      bin_pos[ex.bin] += 1
    elif ex.label_type == datatypes.LabelType.NEGATIVE:
      bin_neg[ex.bin] += 1

  # P(+|b), P(-|b)
  p_pos_bin = {
      b: bin_pos[b] / (bin_pos[b] + bin_neg[b] + 1e-6) for b in bin_weights
  }
  p_neg_bin = {
      b: bin_neg[b] / (bin_pos[b] + bin_neg[b] + 1e-6) for b in bin_weights
  }
  # P(+), P(-) expected value.
  density_ev = np.array(
      [bin_weights[b] * p_pos_bin[b] for b in bin_weights]
  ).sum()
  p_bin_pos = collections.defaultdict(float)
  p_bin_neg = collections.defaultdict(float)
  for b in bin_weights:
    p_bin_pos[b] = bin_weights[b] * p_pos_bin[b] / density_ev
    p_bin_neg[b] = bin_weights[b] * p_neg_bin[b] / (1.0 - density_ev)

  roc_auc = 0

  # For off-diagonal bin pairs:
  # Take the probability of drawing a pos from bin j and neg from bin i.
  # If j > i, all pos examples are scored higher, so contributes directly to the
  # total ROC-AUC.
  bins = sorted(tuple(bin_weights.keys()))
  for i in range(len(bins)):
    for j in range(i + 1, len(bins)):
      roc_auc += p_bin_pos[j] * p_bin_neg[i]

  # For diagonal bin-pairs:
  # Look at actual in-bin observations for diagonal contribution.
  for b in bins:
    bin_pos_scores = np.array([
        v.score
        for v in examples
        if v.bin == b and v.label_type == datatypes.LabelType.POSITIVE
    ])
    bin_neg_scores = np.array([
        v.score
        for v in examples
        if v.bin == b and v.label_type == datatypes.LabelType.NEGATIVE
    ])
    # If either is empty, there's no chance of pulling a (pos, neg) pair from
    # this bin, so we can continue.
    if bin_pos_scores.size == 0 or bin_neg_scores.size == 0:
      continue
    # Count the total number of (pos, neg) pairs where a pos example has a
    # higher score than a negative example.
    hits = (
        (bin_pos_scores[:, np.newaxis] - bin_neg_scores[np.newaxis, :]) > 0
    ).sum()
    bin_roc_auc = hits / (bin_pos_scores.size * bin_neg_scores.size)

    # Contribution is the probability of pulling both pos and neg examples
    # from this bin, multiplied by the bin's ROC-AUC.
    roc_auc += bin_roc_auc * p_bin_pos[b] * p_bin_neg[b]

  return roc_auc



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from perch_hoplite.db import in_mem_impl
from perch_hoplite.db import multi_db_impl

from absl.testing import absltest
from absl.testing import parameterized


class MultiDBImplTest(parameterized.TestCase):

  def test_constructor_empty_wrapper(self):
    with self.assertRaisesRegex(
        ValueError, "MultiDBWrapper contains no databases"
    ):
      multi_db_impl.MultiDBWrapper([])

  def test_split_id_contiguous_indices(self):
    db0 = in_mem_impl.InMemoryGraphSearchDB.create(embedding_dim=1)
    dep_id0 = db0.insert_deployment(name="dep0", project="proj")
    rec_id0 = db0.insert_recording(filename="rec0", deployment_id=dep_id0)
    # Add 10 windows to db0: internal IDs 1-10
    for i in range(10):
      db0.insert_window(recording_id=rec_id0, offsets=[float(i)])

    db1 = in_mem_impl.InMemoryGraphSearchDB.create(embedding_dim=1)
    dep_id1 = db1.insert_deployment(name="dep1", project="proj")
    rec_id1 = db1.insert_recording(filename="rec1", deployment_id=dep_id1)
    # Add 5 windows to db1: internal IDs 1-5
    for i in range(5):
      db1.insert_window(recording_id=rec_id1, offsets=[float(i)])

    wrapper = multi_db_impl.MultiDBWrapper([db0, db1])

    self.assertEqual(wrapper._split_id(2), (0, 1))
    self.assertEqual(wrapper._split_id(4), (0, 2))
    self.assertEqual(wrapper._split_id(5), (1, 2))
    self.assertEqual(wrapper._split_id(7), (1, 3))
    self.assertEqual(wrapper._split_id(14), (0, 7))
    self.assertEqual(wrapper._split_id(16), (0, 8))

    with self.assertRaises(KeyError):
      wrapper._split_id(0)  # db_index is 0, internal_id is 0 (invalid)
    with self.assertRaises(KeyError):
      wrapper._split_id(22)  # db_index is 0, internal_id is 11 (invalid)
    with self.assertRaises(KeyError):
      wrapper._split_id(1)  # db_index is 1, internal_id is 0 (invalid)
    with self.assertRaises(KeyError):
      wrapper._split_id(13)  # db_index is 1, internal_id is 6 (invalid)

  def test_split_id_noncontiguous_indices(self):
    db0 = in_mem_impl.InMemoryGraphSearchDB.create(embedding_dim=1)
    dep_id0 = db0.insert_deployment(name="dep0", project="proj")
    rec_id0 = db0.insert_recording(filename="rec0", deployment_id=dep_id0)
    # Add 10 windows to db0: internal IDs 1-10
    window_ids_0 = []
    for i in range(10):
      window_ids_0.append(
          db0.insert_window(recording_id=rec_id0, offsets=[float(i)])
      )
    # Remove 2 windows from db0: internal IDs 3 and 6
    db0.remove_window(window_ids_0[2])
    db0.remove_window(window_ids_0[5])

    db1 = in_mem_impl.InMemoryGraphSearchDB.create(embedding_dim=1)
    dep_id1 = db1.insert_deployment(name="dep1", project="proj")
    rec_id1 = db1.insert_recording(filename="rec1", deployment_id=dep_id1)
    # Add 5 windows to db1: internal IDs 1-5
    window_ids_1 = []
    for i in range(5):
      window_ids_1.append(
          db1.insert_window(recording_id=rec_id1, offsets=[float(i)])
      )
    # Remove 2 windows from db1: internal IDs 1 and 5
    db1.remove_window(window_ids_1[0])
    db1.remove_window(window_ids_1[4])

    wrapper = multi_db_impl.MultiDBWrapper([db0, db1])

    self.assertEqual(wrapper._split_id(2), (0, 1))
    self.assertEqual(wrapper._split_id(4), (0, 2))
    self.assertEqual(wrapper._split_id(5), (1, 2))
    self.assertEqual(wrapper._split_id(7), (1, 3))
    self.assertEqual(wrapper._split_id(14), (0, 7))
    self.assertEqual(wrapper._split_id(16), (0, 8))

    with self.assertRaises(KeyError):
      wrapper._split_id(6)  # db_index is 0, internal_id is 3 (removed)
    with self.assertRaises(KeyError):
      wrapper._split_id(12)  # db_index is 0, internal_id is 6 (removed)
    with self.assertRaises(KeyError):
      wrapper._split_id(3)  # db_index is 1, internal_id is 1 (removed)
    with self.assertRaises(KeyError):
      wrapper._split_id(11)  # db_index is 1, internal_id is 5 (removed)

  def test_join_id_contiguous_indices(self):
    db0 = in_mem_impl.InMemoryGraphSearchDB.create(embedding_dim=1)
    dep_id0 = db0.insert_deployment(name="dep0", project="proj")
    rec_id0 = db0.insert_recording(filename="rec0", deployment_id=dep_id0)
    # Add 10 windows to db0: internal IDs 1-10
    for i in range(10):
      db0.insert_window(recording_id=rec_id0, offsets=[float(i)])

    db1 = in_mem_impl.InMemoryGraphSearchDB.create(embedding_dim=1)
    dep_id1 = db1.insert_deployment(name="dep1", project="proj")
    rec_id1 = db1.insert_recording(filename="rec1", deployment_id=dep_id1)
    # Add 5 windows to db1: internal IDs 1-5
    for i in range(5):
      db1.insert_window(recording_id=rec_id1, offsets=[float(i)])

    wrapper = multi_db_impl.MultiDBWrapper([db0, db1])

    self.assertEqual(wrapper._join_id(0, 1), 2)
    self.assertEqual(wrapper._join_id(0, 2), 4)
    self.assertEqual(wrapper._join_id(1, 2), 5)
    self.assertEqual(wrapper._join_id(1, 3), 7)
    self.assertEqual(wrapper._join_id(0, 7), 14)
    self.assertEqual(wrapper._join_id(0, 8), 16)

    with self.assertRaises(KeyError):
      wrapper._join_id(0, 0)  # db_index is 0, internal_id is 0 (invalid)
    with self.assertRaises(KeyError):
      wrapper._join_id(0, 11)  # db_index is 0, internal_id is 11 (invalid)
    with self.assertRaises(KeyError):
      wrapper._join_id(1, 0)  # db_index is 1, internal_id is 0 (invalid)
    with self.assertRaises(KeyError):
      wrapper._join_id(1, 6)  # db_index is 1, internal_id is 6 (invalid)

  def test_join_id_noncontiguous_indices(self):
    db0 = in_mem_impl.InMemoryGraphSearchDB.create(embedding_dim=1)
    dep_id0 = db0.insert_deployment(name="dep0", project="proj")
    rec_id0 = db0.insert_recording(filename="rec0", deployment_id=dep_id0)
    window_ids_0 = []
    # Add 10 windows to db0: internal IDs 1-10
    for i in range(10):
      window_ids_0.append(
          db0.insert_window(recording_id=rec_id0, offsets=[float(i)])
      )
    # Remove 2 windows from db0: internal IDs 3 and 6
    db0.remove_window(window_ids_0[2])
    db0.remove_window(window_ids_0[5])

    db1 = in_mem_impl.InMemoryGraphSearchDB.create(embedding_dim=1)
    dep_id1 = db1.insert_deployment(name="dep1", project="proj")
    rec_id1 = db1.insert_recording(filename="rec1", deployment_id=dep_id1)
    window_ids_1 = []
    # Add 5 windows to db1: internal IDs 1-5
    for i in range(5):
      window_ids_1.append(
          db1.insert_window(recording_id=rec_id1, offsets=[float(i)])
      )
    # Remove 2 windows from db1: internal IDs 1 and 5
    db1.remove_window(window_ids_1[0])
    db1.remove_window(window_ids_1[4])

    wrapper = multi_db_impl.MultiDBWrapper([db0, db1])

    self.assertEqual(wrapper._join_id(0, 1), 2)
    self.assertEqual(wrapper._join_id(0, 2), 4)
    self.assertEqual(wrapper._join_id(1, 2), 5)
    self.assertEqual(wrapper._join_id(1, 3), 7)
    self.assertEqual(wrapper._join_id(0, 7), 14)
    self.assertEqual(wrapper._join_id(0, 8), 16)

    with self.assertRaises(KeyError):
      wrapper._join_id(0, 3)  # db_index is 0, internal_id is 3 (removed)
    with self.assertRaises(KeyError):
      wrapper._join_id(0, 6)  # db_index is 0, internal_id is 6 (removed)
    with self.assertRaises(KeyError):
      wrapper._join_id(1, 1)  # db_index is 1, internal_id is 1 (removed)
    with self.assertRaises(KeyError):
      wrapper._join_id(1, 5)  # db_index is 1, internal_id is 5 (removed)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ingest fully-annotated dataset audio and labels."""

import collections
from collections.abc import Callable
import dataclasses

from etils import epath
from ml_collections import config_dict
import pandas as pd
from perch_hoplite.agile import embed
from perch_hoplite.agile import source_info
from perch_hoplite.db import datatypes
from perch_hoplite.db import db_loader
from perch_hoplite.db import interface
from perch_hoplite.taxonomy import annotations_fns
import tqdm

BASE_PATH = epath.Path('gs://chirp-public-bucket/soundscapes')


@dataclasses.dataclass
class AnnotatedDatasetIngestor:
  """Add annotations to embeddings DB from CSV annotations.

  Note that currently we only add positive labels.

  Attributes:
    base_path: Base path for the dataset.
    audio_glob: Glob for the audio files.
    dataset_name: Name of the dataset.
    annotation_filename: Filename for the annotations CSV.
    annotation_load_fn: Function to load the annotations CSV.
  """

  base_path: epath.Path
  audio_glob: str
  dataset_name: str
  annotation_filename: str
  annotation_load_fn: Callable[[str | epath.Path], pd.DataFrame]

  def ingest_dataset(
      self,
      db: interface.HopliteDBInterface,
      provenance: str = 'annotations_csv',
  ) -> collections.defaultdict[str, int]:
    """Load annotations and insert labels into the DB.

    Args:
      db: The DB to insert labels into.
      provenance: The provenance to use for the labels.

    Returns:
      A dictionary of ingested label counts.
    """
    annos_path = epath.Path(self.base_path) / self.annotation_filename
    annos_df = self.annotation_load_fn(annos_path)
    lbl_counts = collections.defaultdict(int)
    file_ids = annos_df['filename'].unique()
    label_set = set()
    for file_id in tqdm.tqdm(file_ids):
      source_annos = annos_df[annos_df['filename'] == file_id]
      for idx in db.match_window_ids(
          deployments_filter=config_dict.create(
              eq=dict(project=self.dataset_name)
          ),
          recordings_filter=config_dict.create(eq=dict(filename=file_id)),
      ):
        window = db.get_window(idx)
        start_offset = window.offsets[0]
        end_offset = window.offsets[1]
        emb_annos = source_annos[source_annos['start_time_s'] < end_offset]
        emb_annos = emb_annos[emb_annos['end_time_s'] > start_offset]
        # All of the remaining annotations match the target embedding.
        for labels in emb_annos['label']:
          for label in labels:
            label_set.add(label)
            db.insert_annotation(
                recording_id=window.recording_id,
                offsets=window.offsets,
                label=label,
                label_type=datatypes.LabelType.POSITIVE,
                provenance=provenance,
                handle_duplicates='allow',
            )
            lbl_counts[label] += 1
    lbl_count = sum(lbl_counts.values())
    print(f'\nInserted {lbl_count} labels.')
    return lbl_counts


CORNELL_LOADER = lambda x: annotations_fns.load_cornell_annotations(
    x, file_id_prefix='audio/'
)


PRESETS: dict[str, AnnotatedDatasetIngestor] = {
    'anuraset': AnnotatedDatasetIngestor(
        base_path=BASE_PATH / 'anuraset',
        dataset_name='anuraset',
        audio_glob='raw_data/*/*.wav',
        annotation_filename='annotations.csv',
        annotation_load_fn=annotations_fns.load_anuraset_annotations,
    ),
    'coffee_farms': AnnotatedDatasetIngestor(
        base_path=BASE_PATH / 'coffee_farms',
        dataset_name='coffee_farms',
        audio_glob='audio/*.flac',
        annotation_filename='annotations.csv',
        annotation_load_fn=CORNELL_LOADER,
    ),
    'hawaii': AnnotatedDatasetIngestor(
        base_path=BASE_PATH / 'hawaii',
        dataset_name='hawaii',
        audio_glob='audio/*.flac',
        annotation_filename='annotations.csv',
        annotation_load_fn=CORNELL_LOADER,
    ),
    'high_sierras': AnnotatedDatasetIngestor(
        base_path=BASE_PATH / 'high_sierras',
        dataset_name='high_sierras',
        audio_glob='audio/*.flac',
        annotation_filename='annotations.csv',
        annotation_load_fn=CORNELL_LOADER,
    ),
    'peru': AnnotatedDatasetIngestor(
        base_path=BASE_PATH / 'peru',
        dataset_name='peru',
        audio_glob='audio/*.flac',
        annotation_filename='annotations.csv',
        annotation_load_fn=CORNELL_LOADER,
    ),
    'powdermill': AnnotatedDatasetIngestor(
        base_path=BASE_PATH / 'powdermill',
        audio_glob='*/*.wav',
        dataset_name='powdermill',
        annotation_filename='powdermill.csv',
        annotation_load_fn=annotations_fns.load_powdermill_annotations,
    ),
    'sierras_kahl': AnnotatedDatasetIngestor(
        base_path=BASE_PATH / 'sierras_kahl',
        dataset_name='sierras_kahl',
        audio_glob='audio/*.flac',
        annotation_filename='annotations.csv',
        annotation_load_fn=CORNELL_LOADER,
    ),
    'ssw': AnnotatedDatasetIngestor(
        base_path=BASE_PATH / 'ssw',
        dataset_name='ssw',
        audio_glob='audio/*.flac',
        annotation_filename='annotations.csv',
        annotation_load_fn=CORNELL_LOADER,
    ),
    'weldy_calltype': AnnotatedDatasetIngestor(
        base_path=BASE_PATH / 'weldy_calltype',
        dataset_name='weldy_calltype',
        audio_glob='annotated_recordings/*.wav',
        annotation_filename='annotations.csv',
        annotation_load_fn=annotations_fns.load_weldy_annotations,
    ),
}


def embed_annotated_dataset(
    ds_choice: str | AnnotatedDatasetIngestor,
    db_path: str,
    db_model_config: embed.ModelConfig,
) -> tuple[interface.HopliteDBInterface, dict[str, int]]:
  """Embed a fully-annotated dataset to SQLite Hoplite DB.

  Args:
    ds_choice: The preset name of the dataset to embed. Alternatively, an
      AnnotatedDatasetIngestor can be provided.
    db_path: The path to the DB.
    db_model_config: The model config for the DB.

  Returns:
    The DB and a dictionary of label counts.
  """
  if isinstance(ds_choice, str):
    ingestor = PRESETS[ds_choice]
  else:
    ingestor = ds_choice

  audio_srcs_config = source_info.AudioSources(
      audio_globs=(
          source_info.AudioSourceConfig(
              dataset_name=ingestor.dataset_name,
              base_path=ingestor.base_path.as_posix(),
              file_glob=ingestor.audio_glob,
              min_audio_len_s=1.0,
              target_sample_rate_hz=-2,
          ),
      )
  )
  db = db_loader.create_new_usearch_db(
      db_path=db_path, embedding_dim=db_model_config.embedding_dim
  )
  print('Initialized DB located at ', db_path)
  worker = embed.EmbedWorker(
      audio_sources=audio_srcs_config, db=db, model_config=db_model_config
  )
  worker.process_all()
  print(f'DB contains {db.count_embeddings()} embeddings.')

  class_counts = ingestor.ingest_dataset(db)
  db.commit()
  return db, class_counts



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Load eBird/Clements labels from source data."""
import typing

from absl import app
from absl import flags
import numpy as np
import pandas as pd
from perch_hoplite.taxonomy import namespace
from perch_hoplite.taxonomy import namespace_db

_SOURCE_FILE = flags.DEFINE_string(
    'source_file', 'source_data/ebird_taxonomy_v2022.csv', 'CSV file to load.'
)
_PREFIX = flags.DEFINE_string(
    'prefix',
    'ebird2022',
    'The prefix to attach to the generated namespaces, class lists, and'
    ' mappings.',
)
_OUTPUT_FILE = flags.DEFINE_string(
    'output_file', 'taxonomy_database.json', 'Output file.'
)

SEABIRD_FAMILIES = {
    'sulidae',
    'fregatidae',
    'stercorariidae',
    'laridae',
    'alcidae',
    'scolopacidae',
}

SEABIRD_ORDERS = {
    'sphenisciformes',
    'procellariiformes',
}


def parse_ebird(
    source_file: str | typing.TextIO, prefix: str
) -> namespace_db.TaxonomyDatabase:
  """Parse an eBird CSV source file.

  This parses the CSV file and generates a taxonomy database containing
  namespaces for all eBird codes and for all species codes. It also contains
  namespaces for the scientific names of the species, genera, families, and
  orders. A separate namespace is created containing all the identifiable
  subspecific groups (ISSFs).

  Mappings are created to map eBird codes to their species, genus, family, and
  order.

  Lastly, a class list is created that contains all seabird species.

  Args:
    source_file: The path or file-like object containing the source file.
    prefix: The prefix to use for all the generated data (e.g., `ebird2021` or
      `ebird2022`, to distinguish between versions).

  Returns:
    A `TaxonomyDatabase` containing all the generated data.
  """

  # Load the CSV data
  df = pd.read_csv(source_file)
  # Lower-case the data
  df = df.applymap(lambda x: x.lower() if isinstance(x, str) else x)
  # Extract the genus from the scientific name
  df['genus'] = df['SCI_NAME'].str.split(' ').str[0]
  # Only keep the scientific family name (ignore the common name)
  df['family'] = df['FAMILY'].str.split(' ').str[0]
  # Correction to spuhs
  df.loc[
      (df['CATEGORY'] == 'spuh')
      & ((df['genus'] == df['ORDER1']) | (df['genus'] == df['family'])),
      'genus',
  ] = np.nan
  # Report species as themselves
  df.loc[df['CATEGORY'] == 'species', 'REPORT_AS'] = df.loc[
      df['CATEGORY'] == 'species', 'SPECIES_CODE'
  ]

  # Namespaces (dictionary key is the name of the namespace)
  namespaces = {
      '': df['SPECIES_CODE'],
      'species': df.loc[df['CATEGORY'] == 'species', 'SPECIES_CODE'],
      'issf': df.loc[df['CATEGORY'] == 'issf', 'SPECIES_CODE'],
      'genera': df['genus'].drop_duplicates().dropna(),
      'families': df['family'].drop_duplicates().dropna(),
      'orders': df['ORDER1'].drop_duplicates().dropna(),
      'clements': df.loc[df['CATEGORY'] == 'species', 'SCI_NAME'],
  }

  # The keys are (mapping name, source namespace, target namespace)
  mappings = {
      ('to_species', '', 'species'): df[
          # Only select rows which should be reported as a species
          df.merge(
              df, left_on='REPORT_AS', right_on='SPECIES_CODE', how='left'
          )['CATEGORY_y']
          == 'species'
      ][['SPECIES_CODE', 'REPORT_AS']],
      ('clements_to_species', 'clements', 'species'): df[
          df['CATEGORY'] == 'species'
      ][['SCI_NAME', 'SPECIES_CODE']],
  }

  for mask, suffix in (
      (df['CATEGORY'] == 'species', 'species'),
      (slice(None), ''),
  ):
    prefix_ = suffix + '_' if suffix else ''
    mappings |= {
        (prefix_ + 'to_genus', suffix, 'genera'): df[mask][
            ['SPECIES_CODE', 'genus']
        ],
        (prefix_ + 'to_family', suffix, 'families'): df[mask][
            ['SPECIES_CODE', 'family']
        ],
        (prefix_ + 'to_order', suffix, 'orders'): df[mask][
            ['SPECIES_CODE', 'ORDER1']
        ],
    }

  if SEABIRD_FAMILIES - set(df['family']):
    raise ValueError('seabird families not found in eBird data')
  if SEABIRD_ORDERS - set(df['ORDER1']):
    raise ValueError('seabird orders not found in eBird data')
  seabirds = df[
      df['family'].isin(SEABIRD_FAMILIES) | df['ORDER1'].isin(SEABIRD_ORDERS)
  ]
  # The keys are class list name, namespace
  class_lists = {
      ('global_seabirds', 'species'): seabirds.loc[
          seabirds['CATEGORY'] == 'species', 'SPECIES_CODE'
      ],
  }

  # Add the prefixes and create the database
  add_prefix = lambda name: (prefix + '_' + name).strip('_')
  namespaces_ = {}
  for name, classes in namespaces.items():
    namespaces_[add_prefix(name)] = namespace.Namespace(frozenset(classes))

  class_lists_ = {}
  for (name, ns), classes in class_lists.items():
    class_lists_[add_prefix(name)] = namespace.ClassList(
        add_prefix(ns), tuple(sorted(classes))
    )

  mappings_ = {}
  for (name, source_ns, target_ns), mapping in mappings.items():
    # Some spuhs don't have a genus, and this was set to nan. Drop these from
    # the mappings.
    mapping = mapping.dropna()
    mappings_[add_prefix(name)] = namespace.Mapping(
        add_prefix(source_ns),
        add_prefix(target_ns),
        dict(zip(mapping.iloc[:, 0], mapping.iloc[:, 1])),
    )

  return namespace_db.TaxonomyDatabase(namespaces_, class_lists_, mappings_)


def main(argv: list[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')
  ebird_db = parse_ebird(_SOURCE_FILE.value, _PREFIX.value)

  # Merge into existing database and write
  db = namespace_db.load_db()
  db.namespaces |= ebird_db.namespaces
  db.mappings |= ebird_db.mappings
  db.class_lists |= ebird_db.class_lists

  with open(_OUTPUT_FILE.value, 'w') as f:
    f.write(namespace_db.dump_db(db))


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tool for reorganizing audio files and annotations for hoplite ingestion."""

import dataclasses
from etils import epath
from perch_hoplite.agile import embed
from perch_hoplite.agile import metadata
from perch_hoplite.agile import source_info
from perch_hoplite.db import db_loader
from perch_hoplite.zoo import model_configs


@dataclasses.dataclass(kw_only=True)
class AudioDatasetIngestor:
  """Ingests premade audio datasets to the hoplite standard format."""

  dataset_name: str
  audio_globs: list[str]
  annotation_filename: str | epath.Path | None = None

  def load_metadata(self, base_path: epath.Path) -> metadata.AgileMetadata:
    """Load metadata and annotations for the dataset."""
    return metadata.AgileMetadata.from_directory(base_path)

  def execute(
      self,
      base_path: epath.Path | str,
      output_path: epath.Path | str | None,
      model_key: str,
      logits_key: str | None = None,
      logits_idxes: tuple[int, ...] | None = None,
      embedding_dim: int | None = None,
  ):
    """Ingest the dataset into a hoplite DB."""
    preset_info = model_configs.get_preset_model_config(model_key)
    db_model_config = embed.ModelConfig(
        model_key=preset_info.model_key,
        embedding_dim=preset_info.embedding_dim,
        model_config=preset_info.model_config,
        logits_key=logits_key,
        logits_idxes=logits_idxes,
    )

    if isinstance(base_path, str):
      base_path = epath.Path(base_path)
    if isinstance(output_path, str):
      output_path = epath.Path(output_path)

    self.load_metadata(base_path)

    if output_path is None:
      output_path = base_path

    audio_srcs_config = source_info.AudioSources(
        audio_globs=tuple([
            source_info.AudioSourceConfig(
                dataset_name=self.dataset_name,
                base_path=base_path.as_posix(),
                file_glob=glob,
                min_audio_len_s=1.0,
                target_sample_rate_hz=-2,
            )
            for glob in self.audio_globs
        ])
    )
    if embedding_dim is None and logits_idxes is not None:
      embedding_dim = len(logits_idxes)
    elif embedding_dim is None and logits_key is not None:
      raise ValueError(
          'Cannot infer embedding dimension from logits key, please specify'
          ' embedding_dim.'
      )
    elif embedding_dim is None:
      embedding_dim = preset_info.embedding_dim
    db = db_loader.create_new_usearch_db(
        db_path=output_path, embedding_dim=embedding_dim
    )
    print('Initialized DB located at ', output_path)
    worker = embed.EmbedWorker(
        audio_sources=audio_srcs_config, db=db, model_config=db_model_config
    )
    worker.process_all()
    print(f'DB contains {db.count_embeddings()} embeddings.')
    return db



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrapper for loading models from KaggleHub with backwards compatibility."""

import kagglehub
import tensorflow as tf

# Older model configs may contain a full model URL, as we used with TFHub.
# These are kept for back-compat, and converted to Kaggle Models slugs.
PERCH_TF_HUB_URL = (
    'https://www.kaggle.com/models/google/'
    'bird-vocalization-classifier/frameworks/TensorFlow2/'
    'variations/bird-vocalization-classifier/versions'
)
PERCH_V2_TF_HUB_URL = (
    'https://www.kaggle.com/models/google/bird-vocalization-classifier/'
    'tensorFlow2/perch_v2'
)
SURFPERCH_TF_HUB_URL = (
    'https://www.kaggle.com/models/google/surfperch/TensorFlow2'
)

PERCH_SLUG = 'google/bird-vocalization-classifier/tensorFlow2/'
BASE_KAGGLE_URL = 'https://www.kaggle.com/models/'

PERCH_V1_SLUG = PERCH_SLUG + 'bird-vocalization-classifier'
PERCH_V2_SLUG = PERCH_SLUG + 'perch_v2'
PERCH_V2_CPU_SLUG = PERCH_SLUG + 'perch_v2_cpu'
SURFPERCH_SLUG = 'google/surfperch/TensorFlow2/1'
HUMPBACK_SLUG = 'google/humpback-whale/tensorFlow2/humpback-whale'
MULTISPECIES_WHALE_SLUG = 'google/multispecies-whale/tensorFlow2/default'
YAMNET_SLUG = 'google/yamnet'
VGGISH_SLUG = 'google/vggish'


def normalize_slug(model_slug: str, model_version: int | None = None) -> str:
  """Convert the old full URLs used by TensorflowHub to KaggleHub slugs."""
  if model_slug == PERCH_TF_HUB_URL:
    return PERCH_V1_SLUG
  elif model_slug == PERCH_V2_TF_HUB_URL:
    return PERCH_V2_SLUG
  elif model_slug == SURFPERCH_TF_HUB_URL:
    return SURFPERCH_SLUG
  elif model_slug.startswith(BASE_KAGGLE_URL):
    return model_slug[len(BASE_KAGGLE_URL) :]

  if model_slug == PERCH_V1_SLUG and model_version in (5, 6, 7):
    # Due to SNAFUs uploading the new model version to KaggleModels,
    # some version numbers were skipped.
    raise ValueError('TFHub version 5, 6, and 7 do not exist.')
  if model_version is not None:
    return f'{model_slug}/{model_version}'
  return model_slug


def load(model_slug: str, model_version: int | None = None):
  """Download and load a model from KaggleHub."""
  if model_slug.startswith('/tmp'):
    # Assume this is a path to a downloaded model, rather than a kaggle model.
    return tf.saved_model.load(model_slug)

  model_path = normalize_slug(model_slug, model_version)
  cached_model_path = kagglehub.model_download(model_path)
  model = tf.saved_model.load(cached_model_path)
  return model


def resolve(model_slug: str, model_version: int | None = None) -> str:
  """Download a model from KaggleHub and return the cached model's path."""
  if model_slug.startswith('/tmp'):
    # Assume this is a path to a downloaded model, rather than a kaggle model.
    return model_slug
  model_path = normalize_slug(model_slug, model_version)
  cached_model_path = kagglehub.model_download(model_path)
  return cached_model_path



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for producing hoplite metadata files from various sources."""

import glob
import os

from etils import epath
import pandas as pd


def convert_wabad_data(
    annotations_csv_path: str, metadata_csv_path: str, output_dir: str
):
  """Converts WABAD pooled annotations to hoplite metadata files.

  Args:
    annotations_csv_path: Path to the WABAD annotations CSV file.
    metadata_csv_path: Path to the WABAD metadata CSV file.
    output_dir: Directory to write hoplite_* files to.
  """
  output_dir = epath.Path(output_dir)
  output_dir.mkdir(parents=True, exist_ok=True)

  print(f'Reading WABAD annotations data from {annotations_csv_path}...')
  # Read CSV with comma separator and dot decimal.
  df = pd.read_csv(annotations_csv_path, sep=',', decimal='.')

  # Create the recording identifier by combining site and filename.
  # This assumes audio files will be accessible like:
  # 'ARD/Recordings/ARD_20211027_072000.wav'
  df['recording_id'] = df['Site'] + '/Recordings/' + df['Recording']

  # 1. Create hoplite_annotations.csv
  print('Generating hoplite_annotations.csv...')
  annotations_df = pd.DataFrame({
      'recording': df['recording_id'],
      'label': df['Species'],
      'start_offset_s': df['Begin_Time_(s)'],
      'end_offset_s': df['End_Time_(s)'],
      'label_type': 'positive',
  })
  annotations_df.to_csv(output_dir / 'hoplite_annotations.csv', index=False)

  # 2. Create hoplite_deployments_metadata.csv
  print(f'Reading WABAD metadata from {metadata_csv_path}...')
  meta_df = pd.read_csv(metadata_csv_path, encoding='latin-1')
  print('Generating hoplite_deployments_metadata.csv...')
  deployment_df = meta_df[[
      'Site ID',
      'Recording location',
      'Biome',
      'Latitude',
      'Longitude',
  ]].copy()
  deployment_df = deployment_df.rename(
      columns={
          'Site ID': 'deployment',
          'Recording location': 'country',
          'Biome': 'biome',
          'Latitude': 'lat',
          'Longitude': 'lon',
      }
  )
  deployment_df = deployment_df.drop_duplicates(
      subset=['deployment']
  ).reset_index(drop=True)
  deployment_df.to_csv(
      output_dir / 'hoplite_deployments_metadata.csv', index=False
  )

  # 3. Create hoplite_metadata_description.csv
  print('Generating hoplite_metadata_description.csv...')
  description_content = (
      'field_name,metadata_level,type,description\n'
      'deployment,deployment,str,"WABAD recording site identifier"\n'
      'country,deployment,str,"Country of recording site"\n'
      'biome,deployment,str,"Biome of recording site"\n'
      'lat,deployment,float,"Latitude of recording site"\n'
      'lon,deployment,float,"Longitude of recording site"\n'
  )
  (output_dir / 'hoplite_metadata_description.csv').write_text(
      description_content
  )

  print(f'\nConversion complete. Files written to {output_dir}')
  print(f'Found {len(annotations_df)} annotations.')
  print(f'Found {len(deployment_df)} unique deployments.')


def convert_anuraset_labels(strong_labels_dir: str, output_csv_path: str):
  """Converts Anuraset strong labels to hoplite_annotations.csv.

  Searches for txt files in strong_labels_dir/*/*.txt, parses them as TSV
  with columns start_s, end_s, label, and converts them to hoplite
  annotations CSV format.

  Args:
    strong_labels_dir: Directory containing the Anuraset strong labels.
    output_csv_path: Path to write the hoplite_annotations.csv file to.
  """
  all_annotations = []
  txt_files = glob.glob(os.path.join(strong_labels_dir, '*/*.txt'))
  print(f'Found {len(txt_files)} label files.')

  def clean_label(label):
    """Removes trailing _[A-Z] from label if present."""
    if (
        isinstance(label, str)
        and len(label) > 2
        and label[-2] == '_'
        and label[-1].isalpha()
        and label[-1].isupper()
    ):
      return label[:-2]
    return label

  for txt_file in txt_files:
    try:
      # Use usecols=[0, 1, 2] to only read the first three columns,
      # ignoring any extra columns caused by extra tabs in a row.
      df = pd.read_csv(
          txt_file,
          sep='\t',
          header=None,
          names=['start_s', 'end_s', 'label'],
          usecols=[0, 1, 2],
      )
    except pd.errors.EmptyDataError:
      print(f'Skipping empty file: {txt_file}')
      continue
    except Exception as e:  # pylint: disable=broad-except
      print(f'Could not parse file {txt_file}: {e}')
      continue

    relative_path = os.path.relpath(txt_file, strong_labels_dir)
    recording_name = os.path.splitext(relative_path)[0] + '.wav'

    for _, row in df.iterrows():
      all_annotations.append({
          'recording': recording_name,
          'label': clean_label(row['label']),
          'start_offset_s': row['start_s'],
          'end_offset_s': row['end_s'],
          'label_type': 'positive',
      })

  if not all_annotations:
    print('No annotations found.')
    return

  annotations_df = pd.DataFrame(all_annotations)
  # Ensure columns are in the right order.
  annotations_df = annotations_df[[
      'recording',
      'label',
      'start_offset_s',
      'end_offset_s',
      'label_type',
  ]]
  annotations_df.to_csv(output_csv_path, index=False)
  print(f'{len(annotations_df)} annotations saved to {output_csv_path}')



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tooling for handling metadata with Hoplite databases."""

import collections
import dataclasses
from typing import Any

from etils import epath
import pandas as pd
from perch_hoplite.db import datatypes


@dataclasses.dataclass
class MetadataField:
  """Describes a metadata field.

  Attributes:
    field_name: The name of the metadata field.
    metadata_level: The level at which metadata applies, either `deployment` or
      `recording`.
    dtype: The data type of the field. Supported types are `str`, `float`,
      `int`, and `bytes`.
    description: An optional description of the field.
  """

  field_name: str
  metadata_level: str
  dtype: str
  description: str = ''

  def cast(self, value: Any) -> Any:
    """Casts a value to the field's dtype."""
    if pd.isna(value):
      return None
    if self.dtype == 'str':
      return str(value)
    if self.dtype == 'float':
      return float(value)
    if self.dtype == 'int':
      return int(value)
    if self.dtype == 'bytes':
      return str(value).encode('utf-8')
    return value


def _cast_df_dtypes(df: pd.DataFrame, fields: dict[str, MetadataField]):
  """Casts dataframe columns to specified dtypes."""
  if not df.empty:
    for field_name in df.columns:
      if field_name in fields:
        field = fields[field_name]
        df[field_name] = df[field_name].apply(field.cast)


@dataclasses.dataclass
class AgileMetadata:
  """Handling of agile metadata from CSV files.

  Attributes:
    deployment_metadata: Metadata for each deployment.
    recording_metadata: Metadata for each recording.
    annotations: Annotations for recordings.
    fields: Description of columns in deployment/recording metadata files.
  """

  deployment_metadata: dict[str, dict[str, Any]] = dataclasses.field(
      default_factory=dict
  )
  recording_metadata: dict[str, dict[str, Any]] = dataclasses.field(
      default_factory=dict
  )
  annotations: collections.defaultdict[str, list[datatypes.Annotation]] = (
      dataclasses.field(default_factory=lambda: collections.defaultdict(list))
  )
  fields: dict[str, MetadataField] = dataclasses.field(default_factory=dict)

  @classmethod
  def from_dataframes(
      cls,
      description_df: pd.DataFrame | None = None,
      deployment_df: pd.DataFrame | None = None,
      recording_df: pd.DataFrame | None = None,
      annotations_df: pd.DataFrame | None = None,
  ) -> 'AgileMetadata':
    """Creates an AgileMetadata instance from pandas dataframes.

    Args:
      description_df: DataFrame with metadata descriptions. Expected columns:
        field_name, metadata_level, type, description.
      deployment_df: DataFrame with deployment metadata.
      recording_df: DataFrame with recording metadata.
      annotations_df: DataFrame with annotations. Expected columns: recording,
        label, start_offset_s, end_offset_s, label_type.

    Returns:
      AgileMetadata instance.
    """
    fields = {}
    deployment_metadata = {}
    recording_metadata = {}
    annotations = collections.defaultdict(list)

    if annotations_df is not None and not annotations_df.empty:
      for _, row in annotations_df.iterrows():
        label_type = None
        for lt in datatypes.LabelType:
          if row['label_type'].lower() == lt.name.lower():
            label_type = lt
            break
        if label_type is None:
          continue
        annotations[row['recording']].append(
            datatypes.Annotation(
                id=-1,
                recording_id=-1,
                label=row['label'],
                offsets=[row['start_offset_s'], row['end_offset_s']],
                label_type=label_type,
                provenance='agile_metadata',
            )
        )

    # If no description, then no metadata to process. Return early.
    if description_df is None or description_df.empty:
      return cls(
          deployment_metadata=deployment_metadata,
          recording_metadata=recording_metadata,
          annotations=annotations,
          fields=fields,
      )

    for _, row in description_df.iterrows():
      fields[row.field_name] = MetadataField(
          field_name=row.field_name,
          metadata_level=row.metadata_level,
          dtype=row.type,
          description=row.description if 'description' in row else '',
      )

    deployment_df = (
        deployment_df if deployment_df is not None else pd.DataFrame()
    )
    recording_df = recording_df if recording_df is not None else pd.DataFrame()

    # Deployment metadata handling.
    if 'deployment' in deployment_df.columns:
      deployment_fields = [
          f for f, v in fields.items() if v.metadata_level == 'deployment'
      ]
      deployment_df = deployment_df[
          [f for f in deployment_fields if f in deployment_df.columns]
      ].copy()
      _cast_df_dtypes(deployment_df, fields)
      deployment_metadata = {
          r['deployment']: r for r in deployment_df.to_dict(orient='records')
      }
    elif not deployment_df.empty:
      raise ValueError(
          'Deployment metadata provided but deployment column missing.'
      )

    # Recording metadata handling.
    if 'recording' in recording_df.columns:
      recording_fields = [
          f for f, v in fields.items() if v.metadata_level == 'recording'
      ]
      recording_df = recording_df[
          [f for f in recording_fields if f in recording_df.columns]
      ].copy()
      _cast_df_dtypes(recording_df, fields)
      recording_metadata = {
          r['recording']: r for r in recording_df.to_dict(orient='records')
      }
    elif not recording_df.empty:
      raise ValueError(
          'Recording metadata provided but recording column missing.'
      )

    return cls(
        deployment_metadata=deployment_metadata,
        recording_metadata=recording_metadata,
        annotations=annotations,
        fields=fields,
    )

  @classmethod
  def from_csv_files(
      cls,
      deployments_path: str | epath.Path,
      recordings_path: str | epath.Path,
      description_path: str | epath.Path,
      annotations_path: str | epath.Path,
  ) -> 'AgileMetadata':
    """Creates an AgileMetadata instance from CSV files."""
    deployments_path = epath.Path(deployments_path)
    recordings_path = epath.Path(recordings_path)
    description_path = epath.Path(description_path)
    annotations_path = epath.Path(annotations_path)

    annotations_df = pd.DataFrame()
    if annotations_path.exists():
      annotations_df = pd.read_csv(
          annotations_path,
          dtype={
              'recording': str,
              'label': str,
              'start_offset_s': float,
              'end_offset_s': float,
              'label_type': str,
          },
      )

    description_df = pd.DataFrame()
    if description_path.exists():
      description_df = pd.read_csv(description_path, index_col=False)

    deployment_df = pd.DataFrame()
    if deployments_path.exists():
      deployment_df = pd.read_csv(deployments_path, index_col=False)
    recording_df = pd.DataFrame()
    if recordings_path.exists():
      recording_df = pd.read_csv(recordings_path, index_col=False)
    return cls.from_dataframes(
        description_df, deployment_df, recording_df, annotations_df
    )

  @classmethod
  def from_directory(cls, metadata_dir: str | epath.Path) -> 'AgileMetadata':
    """Creates an AgileMetadata instance from a directory."""
    metadata_dir = epath.Path(metadata_dir)
    return cls.from_csv_files(
        metadata_dir / 'hoplite_deployments_metadata.csv',
        metadata_dir / 'hoplite_recordings_metadata.csv',
        metadata_dir / 'hoplite_metadata_description.csv',
        metadata_dir / 'hoplite_annotations.csv',
    )

  def get_deployment_metadata(self, deployment: str) -> dict[str, Any]:
    """Returns metadata for the deployment.

    Args:
      deployment: The deployment name to use for matching.

    Returns:
      A dictionary of metadata for the matching deployment, or {}.
      The dictionary will not contain keys where values are None.
    """
    metadata = self.deployment_metadata.get(deployment)
    if metadata:
      return {k: v if not pd.isna(v) else None for k, v in metadata.items()}
    return {}

  def get_recording_metadata(self, recording: str) -> dict[str, Any]:
    """Returns metadata for the recording.

    Args:
      recording: The recording name to use for matching.

    Returns:
      A dictionary of metadata for the matching recording, or {}.
      The dictionary will not contain keys where values are None.
    """
    metadata = self.recording_metadata.get(recording)
    if metadata:
      return {k: v if not pd.isna(v) else None for k, v in metadata.items()}
    return {}

  def get_recording_annotations(
      self, recording: str
  ) -> list[datatypes.Annotation]:
    """Returns annotations for the recording."""
    return self.annotations[recording]
